"""Implement `SolParamHDTDCSSim` class."""
from shamo.core.solutions.parametric import SolParamGetDP
from shamo.hd_tdcs import SolHDTDCSSim
from shamo import DistConstant


class SolParamHDTDCSSim(SolParamGetDP):
    """Store information about a HD-tDCS simulation.

    Parameters
    ----------
    name : str
        The name of the solution.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the solution.

    Other Parameters
    ----------------
    sub_json_paths : list [str]
        The relative paths to the sub-solutions.
    sigmas : dict [str, list [shamo.DistABC, str]]
        The electrical conductivity of the tissues.
    model_json_path : str
        The relative path to the model JSON file.
    references : list [str]
        The names of the references.
    source : list [str]
        The name of the electrode used as a source.
    current : shamo.DistABC
        The injected current.
    """

    def __init__(self, name, parent_path, **kwargs):
        super().__init__(name, parent_path, **kwargs)
        self.update(
            {
                "references": kwargs.get("references", []),
                "source": kwargs.get("source", []),
                "current": kwargs.get("current", DistConstant(0.0)),
            }
        )

    @property
    def sub_class(self):
        return SolHDTDCSSim

    @property
    def references(self):
        """Return the names of the electrodes used as references.

        Returns
        -------
        list [str]
            The names of the electrodes used as references.
        """
        return self["references"]

    @property
    def source(self):
        """Return the name of the electrode used as source.

        Returns
        -------
        str
            The name of the electrode used as source.
        """
        return self["source"][0]

    @property
    def current(self):
        """Return the injected current.

        Returns
        -------
        shamo.DistABC
            The injected current.
        """
        return self["current"]

    def finalize(self, **kwargs):
        """Finalize the solution."""
        self._get_sub_json_paths()
        self.save()



"""Implement the `CompParamValue` class."""
import logging

from shamo.core.problems.single import CompABC
from shamo import DistABC

logger = logging.getLogger(__name__)


class CompParamValue(CompABC):
    """Store information about a random value.

    Other Parameters
    ----------------
    val : shamo.DistABC
        The random value.
    """

    def __init__(self, **kwargs):
        self["val"] = kwargs.get("val", None)

    @property
    def val(self):
        """Return the random value.

        Returns
        -------
        shamo.DistABC
            The random value.
        """
        return self["val"]

    def check(self, name, **kwargs):
        """Check if the value is properly set.

        Parameters
        ----------
        name : str
            The name of the value.

        Raises
        ------
        TypeError
            If the value is not a distribution.
        """
        logger.info(f"Checking parameter value '{name}'.")
        if not isinstance(self["val"], DistABC):
            raise TypeError(f"Specified value for '{name}' is not a distribution.")

    def to_pro_param(self, **kwargs):
        """Return ``None``."""
        return None

    def to_py_param(self, **kwargs):
        """Return ``None``."""
        return None

    def set(self, val):
        """Set the random value.

        Parameters
        ----------
        val : shamo.DistABC
            The random value.
        """
        self["val"] = val

    def to_param(self, name="", **kwargs):
        """Return the fixed and varying parameters separately.

        Parameters
        ----------
        name : str
            The name of the property.

        Returns
        -------
        list [list [str, list [float, str | None]]]
            The fixed parameters.
        list [list [str, list [shamo.DistABC, str | None]]]
            The varying parameters.
        """
        fixed = []
        varying = []
        name = f"{name}.val"
        if self.val.dist_type == "constant":
            return [[name, [self.val.val, None]]], []
        else:
            return [], [[name, [self.val, None]]]



"""Implement `DistUniform` class."""
from .abc import DistABC

import chaospy as chaos


class DistUniform(DistABC):
    """A uniform distribution.

    Parameters
    ----------
    lower : float
        The lower bound of the distribution.
    upper : float
        The upper bound of the distribution.
    """

    def __init__(self, lower, upper):
        super().__init__("uniform")
        self.update({"lower": lower, "upper": upper})

    @property
    def lower(self):
        """Return the lower bound of the distribution.

        Returns
        -------
        float
            The lower bound of the distribution.
        """
        return self["lower"]

    @property
    def upper(self):
        """Return the upper bound of the distribution.

        Returns
        -------
        float
            The upper bound of the distribution.
        """
        return self["upper"]

    @property
    def dist(self):
        """Return the actual distribution.

        Returns
        -------
        chaospy.Uniform
            The actual distribution.
        """
        return chaos.Uniform(self.lower, self.upper)

    @property
    def uniform_dist(self):
        """Return a uniform distribution used for sampling.

        Returns
        -------
        chaospy.Uniform
            The uniform distribution.
        """
        return self.dist

    @property
    def salib_name(self):
        """Return the name of the distribution in SALib.

        Returns
        -------
        str
            The name of the distribution in SALib.
        """
        return "unif"

    @property
    def salib_bounds(self):
        """Return the bounds of the distribution in SALib.

        Returns
        -------
        list [float]
            The bounds of the distribution in SALib.
        """
        return [self.lower, self.upper]



"""Implement the `CompSensor` class."""
import logging

from .abc import CompABC

logger = logging.getLogger(__name__)


class CompSensors(CompABC):
    """Store information about a list of sensors.

    Other Parameters
    ----------------
    sensors : list [str]
        A list of sensor names.
    """

    def __init__(self, **kwargs):
        super().__init__(sensors=kwargs.get("sensors", []))

    def check(self, name, **kwargs):
        """Check if the list of sensors is properly set.

        Parameters
        ----------
        name : str
            The name of the list of sensors.

        Raises
        ------
        RuntimeError
            If an element in the list is not in the model.

        Other Parameters
        ----------------
        sensors : dict [str, shamo.Sensor]
            The sensors of the model.
        """
        logger.info(f"Checking sensors '{name}'.")
        sensors = kwargs.get("sensors", {})
        for s in self["sensors"]:
            if s not in sensors:
                raise RuntimeError(f"Sensor '{s}' not found in model.")

    def to_pro_param(self, **kwargs):
        """Return the sensors required to produce the PRO file.

        Returns
        -------
        list [dict [str, int]]
            The sensors names and physical points.

        Other Parameters
        ----------------
        sensors : dict [str, shamo.Sensor]
            The sensors of the model.
        """
        sensors = kwargs.get("sensors", {})
        return [{"sensor": sensors[s].point.group} for s in self["sensors"]]

    def to_py_param(self, **kwargs):
        """Return the list of sensor names.

        Returns
        -------
        list [str]
            The list of sensor names.
        """
        return self["sensors"]

    def add(self, sensor):
        """Add a sensor to the list.

        Parameters
        ----------
        sensor : str
            The name of the sensor.
        """
        self["sensors"].append(sensor)

    def adds(self, sensors):
        """Add multiple sensors to the list.

        Parameters
        ----------
        sensors : list [str]
            The sensor names.
        """
        self["sensors"].extend(sensors)



"""API for `shamo.core.distributions`."""
from .abc import DistABC
from .constant import DistConstant
from .normal import DistNormal, DistTruncNormal
from .uniform import DistUniform

__all__ = ["DistABC", "DistConstant", "DistNormal", "DistTruncNormal", "DistUniform"]



"""API for `shamo.hd_tdcs`."""
from .simulation.single.problem import ProbHDTDCSSim
from .simulation.single.solution import SolHDTDCSSim

from .simulation.parametric.problem import ProbParamHDTDCSSim
from .simulation.parametric.solution import SolParamHDTDCSSim


__all__ = ["ProbHDTDCSSim", "SolHDTDCSSim", "ProbParamHDTDCSSim", "SolParamHDTDCSSim"]



"""Implement `CompParamTissueProp` class."""
from shamo.core.problems.single import CompTissueProp


class CompParamTissueProp(CompTissueProp):
    """Store information about a random tissue property."""

    def to_param(self, name="", **kwargs):
        """Return the fixed and varying parameters separately.

        Parameters
        ----------
        name : str
            The name of the property.

        Returns
        -------
        list [list [str, list [float, str | None]]]
            The fixed parameters.
        list [list [str, list [shamo.DistABC, str | None]]]
            The varying parameters.
        """
        fixed = []
        varying = []
        for t, p in self.items():
            if name != "":
                t = f"{name}.{t}"
            if p[0].dist_type == "constant":
                fixed.append([t, [p[0].val, p[1]]])
            else:
                varying.append([t, p])
        return fixed, varying



"""Implement `SolParamEEGLeadfield` class."""
from shamo.core.solutions.parametric import SolParamGetDP
from shamo.eeg import SolEEGLeadfield


class SolParamEEGLeadfield(SolParamGetDP):
    """Store information about an EEG leadfield matrix.

    Parameters
    ----------
    name : str
        The name of the solution.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the solution.

    Other Parameters
    ----------------
    sub_json_paths : list [str]
        The relative paths to the sub-solutions.
    sigmas : dict [str, list [shamo.DistABC, str]]
        The electrical conductivity of the tissues.
    model_json_path : str
        The relative path to the model JSON file.
    markers : list [str]
        The names of the markers.
    reference : list [str]
        The name of the reference.
    rois : list [str]
        The names of the tissues of the region of interest.
    sensors : list [str]
        The names of the active sensors.
    shape : tuple [int]
        The shape of the matrix.
    use_grid : bool
        If ``True``, the source space is based on a grid.
    """

    def __init__(self, name, parent_path, **kwargs):
        super().__init__(name, parent_path, **kwargs)
        self.update(
            {
                "markers": kwargs.get("markers", []),
                "reference": kwargs.get("reference", []),
                "rois": kwargs.get("rois", []),
                "sensors": kwargs.get("sensors", []),
                "shape": tuple(kwargs.get("shape", [])),
                "use_grid": kwargs.get("use_grid", False),
            }
        )

    @property
    def sub_class(self):
        return SolEEGLeadfield

    @property
    def shape(self):
        """Return the shape of the matrix.

        Returns
        -------
        tuple [int]
            The shape of the matrix.
        """
        return self["shape"]

    @property
    def n_sensors(self):
        """Return the number of active sensors.

        Returns
        -------
        int
            The number of active sensors.
        """
        return self["shape"][0]

    @property
    def n_sources(self):
        """Return the number of sources.

        Returns
        -------
        int
            The number of sources.
        """
        return self["shape"][1]

    @property
    def markers(self):
        """Return the names of the markers.

        Returns
        -------
        list [str]
            The names of the markers.
        """
        return self["markers"]

    @property
    def reference(self):
        """Return the name of the reference.

        Returns
        -------
        str
            The name of the reference.
        """
        return self["reference"]

    @property
    def sensors(self):
        """Return the names of the active sensors.

        Returns
        -------
        list [str]
            The names of the active sensors.
        """
        return self["sensors"]

    @property
    def rois(self):
        """Return the names of the tissues of the region of interest.

        Returns
        -------
        list [str]
            The names of the tissues of the region of interest.
        """
        return self["rois"]

    @property
    def use_grid(self):
        return self["use_grid"]

    def finalize(self, **kwargs):
        """Finalize the solution."""
        self._get_sub_json_paths()
        sub_sols = self.get_sub_sols()
        sub_sol = sub_sols[0]
        self.update({"shape": sub_sol.shape, "sensors": sub_sol.sensors})
        self.save()



"""Implement `SolParamGetDP` class."""
from .abc import SolParamABC
from shamo.core.distributions import DistABC


class SolParamGetDP(SolParamABC):
    """Store information about the solution of a parametric problem depending on Getdp.

    Parameters
    ----------
    name : str
        The name of the solution.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the solution.

    Other Parameters
    ----------------
    sub_json_paths : list [str]
        The relative paths to the sub-solutions.
    sigmas : dict [str, list [shamo.DistABC, str]]
        The electrical conductivity of the tissues.
    model_json_path : str
        The relative path to the model JSON file.

    See Also
    --------
    shamo.core.solutions.parametric.SolParamABC
    """

    def __init__(self, name, parent_path, **kwargs):
        super().__init__(name, parent_path, **kwargs)
        self.update(
            {
                "sigmas": {
                    t: [DistABC.load(**p[0]), p[1]]
                    for t, p in kwargs.get("sigmas", {}).items()
                },
                "model_json_path": kwargs.get("model_json_path", None),
            }
        )

    @property
    def sigmas(self):
        """Return the electrical conductivity of the tissues.

        Returns
        -------
        dict [str, list [shamo.DistABC, str]]
            The electrical conductivity of the tissues.
        """
        return self["sigmas"]

    @property
    def model_json_path(self):
        """Return the path to the model JSON file.

        Returns
        -------
        pathlib.Path
            The path to the model JSON file.
        """
        return (self.path / self["model_json_path"]).resolve()



"""Implement the `CompTissues` class."""
import logging

from .abc import CompABC

logger = logging.getLogger(__name__)


class CompTissues(CompABC):
    """Store information about a list of tissues.

    Other Parameters
    ----------------
    tissues : list [str]
        The list of tissue names.
    """

    def __init__(self, **kwargs):
        super().__init__(tissues=kwargs.get("tissues", []))

    def check(self, name, **kwargs):
        """Check if the list of tissues is properly set.

        Parameters
        ----------
        name : str
            The name of the list of tissue names.

        Raises
        ------
        RuntimeError
            If a tissue does not exist in the model.

        Other Parameters
        ----------------
        tissues : dict [str, shamo.Tissue]
            The tissues of the model.
        """
        logger.info(f"Checking tissues '{name}'.")
        tissues = kwargs.get("tissues", {})
        for t in self["tissues"]:
            if t not in tissues:
                raise RuntimeError(f"Tissue '{t}' not found in model.")

    def to_pro_param(self, **kwargs):
        """Return the parameters required to generate the PRO file.

        Returns
        -------
        list [dict [str, int]]
            The list of physical groups of the tissues.

        Other Parameters
        ----------------
        tissues : dict [str, shamo.Tissue]
            The tissues of the model.
        """
        tissues = kwargs.get("tissues", {})
        return [{"tissue": tissues[t].vol.group} for t in self["tissues"]]

    def to_py_param(self, **kwargs):
        """Return the parameters required to generate the PY file.

        Returns
        -------
        list [str]
            The list of tissue names.
        """
        return self["tissues"]

    def add(self, tissue):
        """Add a tissue to the list.

        Parameters
        ----------
        tissue : str
            The name of the tissue.
        """
        self["tissues"].append(tissue)

    def adds(self, tissues):
        """Add multiple tissues to the list.

        Parameters
        ----------
        tissues : list [str]
            The names of the tissues.
        """
        self["tissues"].extend(tissues)



"""Implement the `CompTissueProp` class."""
import logging

from .abc import CompABC

logger = logging.getLogger(__name__)


class CompTissueProp(CompABC):
    """Store information about a tissue property."""

    def check(self, name, **kwargs):
        """Check if the tissue property is properly set.

        Parameters
        ----------
        name : str
            The name of the property.

        Raises
        ------
        RuntimeError
            If no property is set for an existing tissue.
            If a field set for a tissue does not exist.

        Other Parameters
        ----------------
        tissues : dict [str, shamo.Tissue]
            The tissues of the model.
        """
        logger.info(f"Checking tissue property '{name}'.")
        tissues = kwargs.get("tissues", {})
        for t in tissues.keys():
            if t not in self:
                raise RuntimeError(f"No property value for tissue '{t}'.")
        for t, p in self.items():
            if p[1] is not None and p[1] not in tissues[t].fields:
                raise RuntimeError(f"No field '{p[1]}' in tissue '{t}'.")

    def to_pro_param(self, name="", **kwargs):
        """Return the parameters required to generate the PRO file.

        Parameters
        ----------
        name : str
            The name of the property.

        Returns
        -------
        list [dict [str, float|str]]
            The tissue properties.

        Other Parameters
        ----------------
        tissues : dict [str, shamo.Tissue]
            The tissues of the model.

        Notes
        -----
        If a field was set for the property of a tissue, its formula is fed with all the
        named parameters and evaluated to produce the expression.
        """
        params = []
        tissues = kwargs.get("tissues", {})
        prop = {name: {t: p[0] for t, p in self.items()}}
        for t, p in self.items():
            if p[1] is None:
                params.append({"tissue": t, "prop": p[0]})
            else:
                params.append(
                    {
                        "tissue": t,
                        "prop": tissues[t].fields[p[1]].gen_formula(**prop, **kwargs),
                    }
                )
        return params

    def to_py_param(self, **kwargs):
        """Return the parameters required to generate the PY file.

        Returns
        -------
        list [dict [str, str|float]]
            The tissue properties.

        Notes
        -----
        Each element of the list is formatted as ``{"tissue": name, "prop": value,
        "field": field_name}``.
        """
        return [{"tissue": t, "prop": p[0], "field": p[1]} for t, p in self.items()]

    def set(self, tissue, prop, field=None):
        """Set the property for a tissue.

        Parameters
        ----------
        tissue : str
            The name of the tissue.
        prop : float
            The value of the property.
        field : str, optional
            The name of the field associated.
        """
        self[tissue] = [prop, field]

    def sets(self, props):
        """Set multiple properties at once.

        Parameters
        ----------
        props : dict [str, float]
            The properties of the tissues.
        """
        for t, p in props.items():
            self[t] = [p, None]



"""Implement the `ComGridSampler` class."""
import logging

import nibabel as nib
import numpy as np
import re

from .abc import CompABC
from shamo.utils.onelab import read_vector_file, pos_to_nii

logger = logging.getLogger(__name__)


class CompGridSampler(CompABC):
    """Store information about a space grid sampler.

    A grid sampler is used to convert outputs from GetDP into a NII file more suitable
    for neuroimaging studies.

    Other Parameters
    ----------------
    affine : numpy.ndarray
        The affine matrix of the grid.
    shape : tuple [int]
            The shape of the grid.
    mask : numpy.ndarray
            The mask of the grid.
    """

    def __init__(self, **kwargs):
        super().__init__(
            affine=kwargs.get("affine", []),
            shape=kwargs.get("shape", []),
            mask=kwargs.get("mask", None),
        )

    @property
    def use_grid(self):
        """Return wether or not the grid can be used.

        Returns
        -------
        bool
            Return ``True`` if the grid can be used, ``False`` otherwise.
        """
        return True if self["affine"] != [] and self["shape"] != [] else False

    @property
    def affine(self):
        """Return the affine matrix of the grid.

        Returns
        -------
        numpy.ndarray
            The affine matrix of the grid.
        """
        return self["affine"]

    @property
    def shape(self):
        """Return the shape of the grid.

        Returns
        -------
        tuple [int]
            The shape of the grid.
        """
        return self["shape"]

    @property
    def mask(self):
        """Return the mask of the grid.

        Returns
        -------
        numpy.ndarray
            The mask of the grid.
        """
        return self["mask"]

    def set(self, affine, shape, mask=None, resize=False):
        """Set the grid component.

        Parameters
        ----------
        affine : numpy.ndarray
            The affine matrix of the grid.
        shame : tuple [int]
            The shape of the grid.
        mask : numpy.ndarray, optional
            The mask of the grid. If set to ``None``, no mask is applied to the
            resulting data. (The default is ``None``)
        resize : bool, optional
            If set to ``True``, the affine matrix is rescaled from meters to
            millimeters. (the default is ``False``)
        """
        self["affine"] = np.vstack((affine[:3, :], np.array([0, 0, 0, 1])))
        if resize:
            self["affine"] = np.diag([1e-3] * 3 + [1]) @ self["affine"]
        self["shape"] = shape
        self["mask"] = mask

    def check(self, name, **kwargs):
        """Check is the grid is properly set.

        Parameters
        ----------
        name : str
            The name of the grid.

        Raises
        ------
        TypeError
            If argument `affine` is not a `numpy.ndarray`.
        ValueError
            If argument `affine` is not (3, 4) or (4, 4) `numpy.ndarray`.
            If argument `shape` is not a 3D shape.
        """
        if self.use_grid:
            logger.info(f"Checking grid sampler '{name}'.")
            if not isinstance(self["affine"], np.ndarray):
                raise TypeError("Argument 'affine' expects type numpy.ndarray.")
            if self["affine"].shape not in ((3, 4), (4, 4)):
                raise ValueError("Argument 'affine' expects shape (3,4) or (4,4).")
            if len(self["shape"]) != 3:
                raise ValueError("Argument 'shape' expects 3 values.")

    def to_pro_param(self, **kwargs):
        """Return if the grid can be used.

        Returns
        -------
        bool
            Return ``True`` if the grid is set.
        """
        return self.use_grid

    def to_py_param(self, **kwargs):
        """Return the parameters required to render python template.

        Returns
        -------
        dict [str, str]
            The parameters required to render python template.
        """
        if self.use_grid:
            return {
                "affine": str(self.affine.tolist()),
                "shape": str(self.shape),
                "mask": str(self.mask.astype(int).tolist())
                if self.mask is not None
                else None,
            }
        else:
            return None

    def nii_from_pos(self, src, dst):
        """Convert a POS file into a NII file.

        Parameters
        ----------
        src : str, byte or os.PathLike
            The path to the POS file to convert.
        dst : str, byte or os.PathLike
            The path where the generated NII image must be saved.
        Returns
        -------
        nibabel.Nifti1Image
            The generated NII image.
        """
        return pos_to_nii(src, dst, self.affine, self.shape, mask=self.mask)



"""Implement `DistConstant` class."""
from .abc import DistABC


class DistConstant(DistABC):
    """A constant value for a random parameter.

    When dealing with parametric problems, some parameters are supposed random. If one
    of them is known, it must be modelled by this class.

    Parameters
    ----------
    val : float
        The constant value.
    """

    def __init__(self, val):
        super().__init__("constant")
        self.update({"val": val})

    @property
    def val(self):
        """Return the constant value.

        Returns
        -------
        float
            The constant value.
        """
        return self["val"]

    @property
    def expect(self):
        """Return the constant value.

        Returns
        -------
        float
            The constant value.
        """
        return self.val

    @property
    def dist(self):
        """Return ``None``."""
        return None

    @property
    def uniform_dist(self):
        """Return ``None``."""
        return None

    @property
    def salib_name(self):
        """Return the name of the distribution in SALib.

        Returns
        -------
        str
            The name of the distribution in SALib.
        """
        return ""

    @property
    def salib_bounds(self):
        """Return the bounds of the distribution in SALib.

        Returns
        -------
        list [float]
            The bounds of the distribution in SALib.
        """
        return []



import os
import subprocess as sub
import yaml

import click
import gmsh


def _get_os_info():
    info = os.uname()
    return {
        "os": info.sysname,
        "release": info.release,
        "version": info.version,
        "architecture": info.machine,
    }


def _get_gmsh_info():
    gmsh.initialize()
    info = gmsh.option.getString("General.BuildInfo")
    gmsh.finalize()
    lines = info.split("; ")
    pairs = [l.split(": ") for l in lines]
    info = {p[0].replace(" ", "-").lower(): p[1] for p in pairs}
    unwanted = ["license", "build-host", "web-site", "issue-tracker"]
    for u in unwanted:
        info.pop(u, None)
    info["build-options"] = info["build-options"].split(" ")
    return info


def _get_getdp_info():
    info = sub.run(["getdp", "-info"], capture_output=True).stderr.decode("utf-8")
    lines = info.strip().split("\n")
    pairs = [l.split(": ") for l in lines]
    info = {p[0].strip().replace(" ", "-").lower(): p[1] for p in pairs}
    unwanted = ["license", "build-host", "web-site", "issue-tracker"]
    for u in unwanted:
        info.pop(u, None)
    build = ["build-options", "gmsh-lib-options"]
    for b in build:
        info[b] = info[b].split(" ")
    return info


def _get_pip_info():
    info = sub.run(["pip", "list"], capture_output=True).stdout.decode("utf-8")
    lines = info.strip().split("\n")[2:]
    pairs = [l.split() for l in lines]
    return {p[0]: p[1] for p in pairs}


@click.command()
def main():
    _get_os_info()
    report = {
        "os": _get_os_info(),
        "gmsh": _get_gmsh_info(),
        "getdp": _get_getdp_info(),
        "packages": _get_pip_info(),
    }
    click.echo(yaml.dump(report))



"""Implement the `ProbGetDP` class."""
from abc import abstractproperty, abstractmethod
import logging
from pathlib import Path
import re
from subprocess import CalledProcessError, Popen, PIPE, STDOUT

from jinja2 import Environment, PackageLoader
import numpy as np

from .abc import ProbABC
from .components.tissue_property import CompTissueProp

logger = logging.getLogger(__name__)


class ProbGetDP(ProbABC):
    """A base class for any GetDP problem.

    Attributes
    ----------
    sigmas : shamo.core.problems.single.CompTissueProp
        The electrical conductivity of the tissues.

    See Also
    --------
    shamo.core.problems.single.ProbABC
    """

    def __init__(self, **kwargs):
        self._vol = CompTissueProp()
        self.sigmas = CompTissueProp()

    @abstractproperty
    def template(self):
        """Return the name of the template PRO file.

        Returns
        -------
        str
            The name of the template PRO file.
        """

    def _prepare_pro_file_params(self, **kwargs):
        """Return a dict filled with all the required parameters to generate a PY file.

        Returns
        -------
        dict [str, Any]
            The dict filled with all the required parameters.
        """
        return {
            "vol": self._vol.to_pro_param(**kwargs),
            "sigmas": self.sigmas.to_pro_param(name="sigma", **kwargs),
        }

    def _check_components(self, **kwargs):
        """Check if all the components are properly set."""
        self._vol.check("regions", **kwargs)
        self.sigmas.check("sigmas", **kwargs)

    @abstractmethod
    def _prepare_py_file_params(self, **kwargs):
        """Return a dict filled with all the required parameters to generate a PRO file.

        Returns
        -------
        dict [str, Any]
            The dict filled with all the required parameters.
        """
        return {"sigmas": self.sigmas.to_py_param(**kwargs)}

    def _gen_pro_file(self, tmp_dir, **kwargs):
        """Generate the PRO file based on the template.

        Parameters
        ----------
        tmp_dir : str, byte or os.PathLike
            The path to the directory the PRO file must be created in.
        """
        logger.info("Generating problem file.")
        env = Environment(loader=PackageLoader("shamo", "templates/pro"))
        template = env.get_template(self.template)
        content = template.render(**self._prepare_pro_file_params(**kwargs))
        logger.debug(content)
        with open(Path(tmp_dir) / "problem.pro", "w") as f:
            f.write(content)

    def _run_getdp(self, model, tmp_dir):
        """Run GetDP to solve the PRO file.

        Parameters
        ----------
        model : shamo.fem.FEM
            The model to solve the problem for.
        tmp_dir : str, byte or os.PathLike
            The path to the directory the PRO file must be created in.

        Raises
        ------
        CalledProcessError
            If GetDP exits with a wrong exit code.
        """
        cmd = [
            "getdp",
            "problem.pro",
            "-msh",
            str(model.mesh_path.resolve()),
            "-solve",
            "res",
            "-bin",
            "-v2",
        ]
        logger.info(f"Running GetDP with command: {' '.join(cmd)}")
        process = Popen(cmd, stdout=PIPE, stderr=STDOUT, cwd=tmp_dir)
        prev_line = ""
        n_lines = 1
        with process.stdout as pipe:
            for l in iter(pipe.readline, b""):
                if l == prev_line:
                    if n_lines % 1000 == 0:
                        logger.info(f"... (x{n_lines})")
                    n_lines += 1
                if l != prev_line:
                    if n_lines > 1:
                        logger.info(f"{prev_line.decode('utf-8').strip()} (x{n_lines})")
                    logger.info(re.sub(r"[A-Z][a-z ]+: ", "", l.decode().strip()))
                    n_lines = 1
                prev_line = l
        exitcode = process.wait()
        if exitcode != 0:
            raise CalledProcessError(exitcode, cmd)



"""Implement `ProbParamHDTDCSSim` class."""
import logging

from shamo.core.problems.parametric import ProbParamGetDP, CompParamValue
from shamo.core.problems.single import CompGridSampler, CompSensors
from shamo.hd_tdcs import ProbHDTDCSSim
from .solution import SolParamHDTDCSSim

logger = logging.getLogger(__name__)


class ProbParamHDTDCSSim(ProbParamGetDP):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.grid = CompGridSampler()
        self.references = CompSensors()
        self.source = CompSensors()
        self.current = CompParamValue()

    @property
    def template(self):
        """Return the name of the template PY file.

        Returns
        -------
        str
            The name of the template PY file.
        """
        return "hdtdcs_simulation.tmplt"

    def solve(
        self,
        name,
        parent_path,
        model,
        n_evals,
        method="sequential",
        n_proc=1,
        skip=0,
        **kwargs,
    ):
        """Simulate a parametric HD-tDCS.

        Parameters
        ----------
        name : str
            The name of the solution.
        parent_path : str, byte or os.PathLike
            The path to the parent directory of the solution.
        model : shamo.fem.FEM
            The finite element model to solve the problem for.
        n_evals : int
            The number of evaluation points.
        method : str, optional
            The method to solve the sub-problems. The accepted values are
            ``'sequential'``, ``'multiprocessing'``, ``'job'``. (The default is
            ``'sequential'``)
        n_proc : int, optional
            The number of processes to solve the problem when `method` is set to
            ``'multiprocessing'``. (The default is ``1``)
        skip : int, optional
            The number of points to skip at the beginning of the sequence. (The default
            is ``0``)

        Returns
        -------
        shamo.hd_tdcs.SolParamHDTDCSSim
            The parametric HD-tDCS solution.
        """
        logger.info("Solving problem")
        self._check_components(**model, **kwargs)
        params = self._gen_params(n_evals, skip, **kwargs)
        sub_probs = self._gen_sub_probs(n_evals, **params, **kwargs)
        sol = SolParamHDTDCSSim(
            name,
            parent_path,
            references=self.references["sensors"],
            source=self.source["sensors"][0],
            current=self.current["val"],
            sigmas=self.sigmas,
        )
        sol["model_json_path"] = str(sol.get_relative_path(model.json_path))
        sub_sols = self._solve_sub_probs(
            sol, sub_probs, method=method, n_proc=n_proc, model=model, **kwargs
        )
        return sol

    def _gen_fixed_varying(self, **kwargs):
        """Generate two lists containing the fixed and varying parameters.

        Returns
        -------
        list [dict {str, ...}]
            The fixed parameters.
        list [dict {str, ...}]
            The varying parameters.
        """
        fix_sigmas, var_sigmas = self.sigmas.to_param("sigmas", **kwargs)
        fix_current, var_current = self.current.to_param("current", **kwargs)
        return [*fix_sigmas, *fix_current], [*var_sigmas, *var_current]

    def _check_components(self, **kwargs):
        """Check if the components are properly set."""
        self.grid.check("grid", **kwargs)
        self.source.check("source", **kwargs)
        self.references.check("references", **kwargs)
        self.current.check("current", **kwargs)

    def _gen_sub_prob(self, i, **kwargs):
        """Generate one sub-problem.

        Parameters
        ----------
        i : int
            The index of the sub-problem.

        Returns
        -------
        shamo.hd_tdcs.ProbHDTDCSSim
            The generated sub-problem.

        Other Parameters
        ----------------
        sigmas : dict [str, list[numpy.ndarray, str]]
            The electrical conductivity of the tissues.
        """
        s = kwargs.get("sigmas", {})
        prob = ProbHDTDCSSim()
        for t, p in s.items():
            prob.sigmas.set(t, p[0][i], p[1])
        prob.references = self.references
        prob.source = self.source
        c = kwargs.get("current", {})
        prob.current = c["val"][0][i]
        prob.grid = self.grid
        return prob



"""Implement the `ObjDir` class."""
import json
from pathlib import Path

from .abc import ObjABC
from shamo.utils.path import get_relative_path as get_rel_path


class ObjDir(ObjABC):
    """A base class for any savable/loadable object stored as a directory.

    Parameters
    ----------
    name : str
        The name of the object.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the object.

    See Also
    --------
    shamo.core.ObjABC
    """

    def __init__(self, name, parent_path):
        super().__init__(name, parent_path)
        self.path.mkdir(parents=True, exist_ok=True)

    @property
    def path(self):
        """Return the path to the object directory.

        Returns
        -------
        pathlib.Path
            The path to the object directory.
        """
        return self.parent_path / self.name

    @property
    def json_path(self):
        """Return the path to the object JSON file.

        Returns
        -------
        pathlib.Path
            The path to the object JSON file.
        """
        return self.path / f"{self.name}.json"

    @classmethod
    def _split_json_path(cls, json_path):
        """Return the path and the name of the object from the path of the JSON file.

        Parameters
        ----------
        json_path : pathlib.Path
            The path to the JSON file containing the object data.

        Returns
        -------
        str
            The name of the object.
        pathlib.Path
            The path to the parent directory of the object.
        """
        return json_path.stem, json_path.parents[1]

    def get_relative_path(self, path):
        """Return the relative path from the object to a file or directory.

        Parameters
        ----------
        path : str, byte or os.PathLike
            The path to the file or directory to compute the relative path for.

        Returns
        -------
        pathlib.Path
            The relative path to the file or directory.
        """
        return get_rel_path(self.path, path)



"""Implement the `ProbABC` class."""
from abc import ABC, abstractmethod


class ProbABC(dict, ABC):
    """A base class for any problem.

    All the main processes are defined as problem/solution pairs. The problem defines
    the parameters and builds the solution.
    """

    @abstractmethod
    def _check_components(self, **kwargs):
        """Check if all the components are properly set."""

    @abstractmethod
    def _prepare_py_file_params(self, **kwargs):
        """Return a dict filled with all the required parameters to generate a PY file.

        Returns
        -------
        dict [str, Any]
            The dict filled with all the required parameters.
        """

    @abstractmethod
    def solve(self, name, parent_path, **kwargs):
        """Solve the problem and build the solution object.

        Parameters
        ----------
        name : str
            The name of the solution.
        parent_path : str, byte or os.PathLike
            The path to the parent directory of the solution.

        Returns
        -------
        shamo.core.objects.ObjDir
            The solution.
        """



"""Implement the `Sensor` class."""


class Sensor(dict):
    """A FEM sensor.

    Parameters
    ----------
    tissue : str
        The tissue the sensor is in.
    sensor_type : str
        The type of the sensor.
    """

    TYPE_POINT = "point"

    def __init__(self, tissue, sensor_type):
        super().__init__({"tissue": tissue, "sensor_type": sensor_type})

    @property
    def tissue(self):
        """Return the tissue the sensor is in.

        Returns
        -------
        str
            The tissue the sensor is in.
        """
        return self["tissue"]

    @property
    def sensor_type(self):
        """Return the type of the sensor.

        Returns
        -------
        str
            The type of the sensor.
        """
        return self["sensor_type"]



"""Implement the `PointSensor` class."""
from collections.abc import Iterable

from shamo.core.fem import Sensor, Group


class PointSensor(Sensor):
    """A FEM sensor.

    Parameters
    ----------
    tissue : str
        The tissue the sensor is in.
    real_coords : Iterable [float]
        The coordinates of the sensor in the real world.
    mesh_coords : Iterable [float]
        The coordinates of the sensor in the mesh.
    point : shamo.fem.Group
        The physical group of the sensor.
    node : int
        The tag of the node corresponding to the sensor.
    """

    def __init__(self, tissue, real_coords, mesh_coords, point, node, **kwargs):
        super().__init__(tissue, Sensor.TYPE_POINT)
        self.update(
            {
                "real_coords": tuple(real_coords),
                "mesh_coords": tuple(mesh_coords),
                "point": Group(**point),
                "node": int(node),
            }
        )

    @property
    def real_coords(self):
        """Return the real coordinates of the sensor.

        Returns
        -------
        tuple (float, float, float)
            The real coordinates of the sensor [m].
        """
        return self["real_coords"]

    @property
    def mesh_coords(self):
        """Return the mesh coordinates of the sensor.

        Returns
        -------
        tuple (float, float, float)
            The mesh coordinates of the sensor [m].
        """
        return self["mesh_coords"]

    @property
    def point(self):
        """Return the group of the sensor.

        Returns
        -------
        shamo.fem.Group
            The group of the sensor.
        """
        return self["point"]

    @property
    def node(self):
        """Return the node tag of the sensor.

        Returns
        -------
        int
            The node tag of the sensor.
        """
        return self["node"]



"""Implement `SolParamGetDP` class."""
from shamo.core.problems.parametric import ProbParamABC
from .components.tissue_property import CompParamTissueProp
from shamo.utils.path import get_relative_path


class ProbParamGetDP(ProbParamABC):
    """A base class for any parametric problem depending on GetDP."""

    def __init__(self, **kwargs):
        self.sigmas = CompParamTissueProp()

    def _gen_py_file(self, prob, name, parent_path, kwargs):
        """Generate a PY file to solve a sub-problem later.

        Parameters
        ----------
        prob : shamo.core.problems.single.ProbABC
            The problem to solve.
        name : str
            The name of the sub-problem.
        parent_path : str, byte or os.PathLike
            The path to the parent directory of the sub-problem.
        kwargs : dict[str, ...]
            Any other arguments.

        Returns
        -------
        list []

        See Also
        --------
        shamo.core.solutions.parametric.SolParam._gen_py_file
        """
        kwargs["model_path"] = str(
            get_relative_path(parent_path, kwargs.get("model", None).json_path)
        )
        return super()._gen_py_file(prob, name, parent_path, kwargs)

    def _solve_sub_probs(
        self, sol, sub_probs, model, method=ProbParamABC.METHOD_SEQ, n_proc=1, **kwargs,
    ):
        """Solve the sub-problems.

        Parameters
        ----------
        sol : shamo.core.solutions.parametric.SolParamABC
            The parametric solution which the sub-solution will be part of.
        sub_prob : shamo.core.problems.single.ProbABC
            The sub-problems to solve.
        model : shamo.FEM
            The finite element model to solve the problem for.
        method : str, optional
            The method to solve the sub-problems. The accepted values are
            ``'sequential'``, ``'multiprocessing'``, ``'job'``. (The default is
            ``'sequential'``)
        n_proc : int, optional
            The number of processes to solve the problem when `method` is set to
            ``'multiprocessing'``. (The default is ``1``)

        See Also
        --------
        shamo.core.solutions.parametric.SolParam._solve_sub_probs

        Notes
        -----
        If `method` is set to ``'job'``, the method will only produce PY files. Each of
        these scripts will solve a sub-problem. Once all the sub-problems are solved,
        one must call `shamo.core.solutions.parametric.SolABC.finalize()` to finalize
        the resolution.
        """
        kwargs["model"] = model
        return super()._solve_sub_probs(sol, sub_probs, method, n_proc, **kwargs)



"""API for `shamo.utils.onelab`."""
from pathlib import Path

import gmsh
import nibabel as nib
import numpy as np
from scipy.spatial.distance import cdist


def read_vector_file(path):
    """Read a table formatted vector field from GetDP.

    Parameters
    ----------
    path : str, byte or os.PathLike
        The path to the output file.

    Returns
    -------
    int
        The data type.
    np.ndarray
        The element tags.
    np.ndarray
        The values.
    """
    data = np.loadtxt(
        Path(path),
        dtype={
            "names": ("type", "tag", "x", "y", "z"),
            "formats": (np.uint8, np.uint, np.float, np.float, np.float),
        },
        usecols=(0, 1, -3, -2, -1),
    )
    return (
        np.unique(data["type"])[0],
        data["tag"],
        np.vstack((data["x"], data["y"], data["z"])).T,
    )


def get_elems_coords(elem_type, elems_tags):
    """Get the coordinates of specified elements.

    Parameters
    ----------
    elem_type : int
        The type of the elements in `elem_tags`.
    elems_tags : numpy.ndarray
        The tags of the elements to acquire the coordinates for.

    Returns
    -------
    numpy.ndarray
        The coordinates of the specified elements.

    Notes
    -----
    This method must be called inside a Gmsh context.
    """
    coords = []
    tags = []
    coords = gmsh.model.mesh.getBarycenters(elem_type, -1, False, False)
    coords = coords.reshape((-1, 3))
    tags = gmsh.model.mesh.getElementsByType(elem_type, -1)[0]
    idx = np.argsort(tags)
    tags = tags[idx]
    coords = coords[idx, :]
    idx = elems_tags - tags[0]
    coords = coords[idx, :]
    return coords


def get_elems_subset(dim, tags, min_dist):
    """Get a subset of elements based on a minimal distance between them.

    Parameters
    ----------
    dim : int
        The dimension of the elements.
    tags : list [int]
        The entity tags.
    min_dist : float
        The minimal distance between elements.

    Returns
    -------
    numpy.ndarray
        The elements tags.
    numpy.ndarray
        The elements coordinates.
    """
    elems_tags = []
    for t in tags:
        elems_type, ts, _ = gmsh.model.mesh.getElements(dim, t)
        elems_tags.append(ts)
    elems_tags = np.hstack(elems_tags)
    elems_type = elems_type[0]
    elems_tags = elems_tags[0]
    coords = get_elems_coords(elems_type, elems_tags)
    sub_elems_tags = [elems_tags[0]]
    sub_elems_coords = [coords[0]]
    while elems_tags.size > 0:
        dist = cdist([sub_elems_coords[-1]], coords)[0]
        idx = dist < min_dist
        elems_tags = np.delete(elems_tags, idx)
        coords = np.delete(coords, idx, axis=0)
        dist = np.delete(dist, idx)
        if elems_tags.size > 0:
            idx = np.argmin(dist)
            sub_elems_tags.append(elems_tags[idx])
            sub_elems_coords.append(coords[idx])
    return sub_elems_tags, sub_elems_coords


def pos_to_nii(src, dst, affine, shape, mask=None):
    """Convert a POS file into a NII file.

    Parameters
    ----------
    src : str, byte or os.PathLike
        The path to the input NII file.
    dst : str, byte or os.PathLike
        The path to the output NII file.
    affine : numpy.ndarray
        The affine matrix of the NII volume.
    shape : Iterable [float]
        The shape of the NII volume.
    mask : numpy.ndarray, optional
        If not set to ``None``, the mask is applied to the field.

    Returns
    -------
    nibabel.Nifti1Image
        The generated NII image.
    """
    # Define axis
    o = affine @ np.array([-0.5, -0.5, -0.5, 1]).T
    u = affine @ np.array([shape[0] - 0.5, -0.5, -0.5, 1]).T
    v = affine @ np.array([-0.5, shape[1] - 0.5, -0.5, 1]).T
    w = affine @ np.array([-0.5, -0.5, shape[2] - 0.5, 1]).T
    gmsh.initialize()
    gmsh.option.setNumber("General.Terminal", 1)
    gmsh.option.setNumber("General.Verbosity", 5)
    gmsh.option.setNumber("Mesh.Binary", 1)
    gmsh.merge(str(Path(src)))
    # Set plugin parameters
    gmsh.plugin.setNumber("CutBox", "X0", o[0])
    gmsh.plugin.setNumber("CutBox", "Y0", o[1])
    gmsh.plugin.setNumber("CutBox", "Z0", o[2])
    gmsh.plugin.setNumber("CutBox", "X1", u[0])
    gmsh.plugin.setNumber("CutBox", "Y1", u[1])
    gmsh.plugin.setNumber("CutBox", "Z1", u[2])
    gmsh.plugin.setNumber("CutBox", "X2", v[0])
    gmsh.plugin.setNumber("CutBox", "Y2", v[1])
    gmsh.plugin.setNumber("CutBox", "Z2", v[2])
    gmsh.plugin.setNumber("CutBox", "X3", w[0])
    gmsh.plugin.setNumber("CutBox", "Y3", w[1])
    gmsh.plugin.setNumber("CutBox", "Z3", w[2])
    gmsh.plugin.setNumber("CutBox", "Boundary", 0)
    gmsh.plugin.setNumber("CutBox", "NumPointsU", shape[0] + 1)
    gmsh.plugin.setNumber("CutBox", "NumPointsV", shape[1] + 1)
    gmsh.plugin.setNumber("CutBox", "NumPointsW", shape[2] + 1)
    # Run plugin
    gmsh.plugin.run("CutBox")
    # Get field data
    dtype, n_elems, data = gmsh.view.getListData(1)
    gmsh.view.write(1, str(Path(dst).parent / f"{Path(dst).stem}_grid.pos"))
    gmsh.finalize()
    n_elems = n_elems[0]
    data = data[0]
    data = data.reshape((n_elems, -1))
    n_vals = int(data[:, 24:].shape[1] / 8)
    vals = np.zeros((n_elems, n_vals))
    for i in range(n_vals):
        vals[:, i] = np.mean(data[:, 24 + i :: n_vals], axis=1)
    field = vals.reshape((*shape, n_vals))
    if mask is not None:
        field[~mask, ...] = 0
    # Write NII file
    img = nib.Nifti1Image(field, affine)
    img.to_filename(Path(dst))
    return img



"""Implement the `SurrABC` class."""
from abc import abstractclassmethod
import pickle

from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern, ConstantKernel
from sklearn.multioutput import MultiOutputRegressor

from shamo.core.objects import ObjDir
from shamo import DistABC


class SurrABC(ObjDir):
    """Generate a Gaussian process from a set of training data.

    Parameters
    ----------
    name : str
        The name of the generated surrogate model.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the surrogate model.

    Other Parameters
    ----------------
    params : list [tuple [str, shamo.DistABC]]
        A list of tuples containing the names of the parameters and the
        corresponding distributions as values.
    sol_json_path : str
        The path to the parametric solution the surrogate is built of.
    """

    def __init__(self, name, parent_path, **kwargs):
        super().__init__(name, parent_path)
        self.update(
            {
                "params": [[n, DistABC.load(**d)] for n, d in kwargs.get("params", [])],
                "sol_json_path": kwargs.get("sol_json_path", None),
            }
        )

    @property
    def gp_path(self):
        """Return the path of the Gaussian process file.

        Returns
        -------
        pathlib.Path
            The path of the Gaussian process file.
        """
        return self.path / f"{self.name}.gp"

    @property
    def sol_json_path(self):
        return (self.path / self["sol_json_path"]).resolve()

    @property
    def params(self):
        """Return the parameters of the surrogate model.

        Returns
        -------
        list [tuple [str, shamo.DistABC]]
            A list of tuples containing the names of the parameters and the
            corresponding distributions as values.
        """
        return self["params"]

    @abstractclassmethod
    def _check_params(cls, **kwargs):
        """Check if the parameters are properly set.

        Notes
        -----
        This method must be implemented to be able to generate a surrogate model.
        """

    @abstractclassmethod
    def _get_data(cls, sol, **kwargs):
        """Extract relevant data from a parametric solution.

        Parameters
        ----------
        sol : shamo.core.solutions.parametric.SolParamABC
            The parametric solution to generate a surrogate model for.

        Returns
        -------
        numpy.ndarray
            The coordinates of the training points in the parameter space. Each row
            represents an evaluation.
        numpy.ndarray
            The observations of the actual model at each coordinate from `x`. Each row
            represents an observation.
        list [tuple [str, shamo.DistABC]]
            A list of tuples containing the names of the parameters and the
            corresponding distributions as values.

        Notes
        -----
        This method must be implemented to be able to generate a surrogate model.
        """

    @classmethod
    def fit(cls, name, parent_path, sol, **kwargs):
        """Generate a Gaussian process from a set of training data.

        Parameters
        ----------
        name : str
            The name of the generated surrogate model.
        parent_path : str, byte or os.PathLike
            The path to the parent directory of the surrogate model.
        sol : shamo.core.solutions.parametric.SolParamABC
            The parametric solution to generate a surrogate model for.

        Returns
        -------
        shamo.Surrogate
            The generated surrogate model.

        Other Parameters
        ----------------
        kernel : sklearn.gaussian_process.kernels.Kernel
            The kernel used to generate the Gaussian process.
        n_restarts_optimizer : int
            The number of restarts for the optimisation step.
        random_state : int
            The seed for the random state.
        alpha : float
            The added diagonal to account for noise on training points.
        n_proc : int
            The number of jobs to run in parallel. ``None`` means 1 job runs at a time
            and ``-1`` means all cores are used.

        See Also
        --------
        sklearn.gaussian_process.GaussianProcessRegressor
        """
        cls._check_params(**kwargs)
        x, y, params = cls._get_data(sol, **kwargs)
        kernel = kwargs.get(
            "kernel",
            ConstantKernel() * Matern(length_scale=[1.0] * len(params), nu=2.5),
        )
        n_restarts_optimizer = kwargs.get("n_restarts_optimizer", 0)
        random_state = kwargs.get("random_state", 0)
        alpha = kwargs.get("alpha", 1e-10)
        if y.ndim > 1 and y.shape[1] > 1:
            gp = MultiOutputRegressor(
                GaussianProcessRegressor(
                    kernel=kernel,
                    n_restarts_optimizer=n_restarts_optimizer,
                    random_state=random_state,
                    normalize_y=True,
                    alpha=alpha,
                ),
                n_jobs=kwargs.get("n_proc", None),
            ).fit(x, y)
        else:
            gp = GaussianProcessRegressor(
                kernel=kernel,
                n_restarts_optimizer=n_restarts_optimizer,
                random_state=random_state,
                normalize_y=True,
                alpha=alpha,
            ).fit(x, y)
        surr = cls(name, parent_path, params=params)
        surr["sol_json_path"] = str(surr.get_relative_path(sol.json_path))
        pickle.dump(gp, open(surr.gp_path, "wb"))
        surr.save()
        return surr

    def get_gp(self):
        """Load the Gaussian process.

        Returns
        -------
        sklearn.gaussian_process.GaussianProcessRegressor
            The Gaussian process.
        """
        return pickle.load(open(self.gp_path, "rb"))

    def predict(self, x, **kwargs):
        """Evaluate the Gaussian process on new points.

        Parameters
        ----------
        x : numpy.ndarray
            The coordinates of the evaluation points in the parameter space. Each row
            represents an evaluation.

        Notes
        -----
        To change the behaviour of this method, one should subclass `Surrogate` and
        change the `_post_pro` method.
        """
        gp = self.get_gp()
        # MultiOutputRegressor does not pass any parameter to each estimator so it is
        # not possible to obtain the standard deviation for this type of regressor.
        if isinstance(gp, MultiOutputRegressor):
            y_mean = gp.predict(x)
            y_std = None
        else:
            y_mean, y_std = gp.predict(x, return_std=True)
        return self._post_pro(x, y_mean, y_std, **kwargs)

    def _post_pro(self, x, y_mean, y_std, **kwargs):
        """Applies a post-processing operation to the predictions.

        Parameters
        ----------
        numpy.ndarray
            The evaluation points.
        numpy.ndarray
            The evaluations mean.
        numpy.ndarray
            The evaluations standard deviations.

        Notes
        -----
        This method should be modified to apply post-processing operations.
        """
        return y_mean, y_std



"""Implement `DistABC` class."""
from abc import ABC, abstractproperty

import chaospy as chaos


class DistABC(dict, ABC):
    """A base class for any probability distribution.

    Parameters
    ----------
    dist_type : str
        The type of the distribution.
    """

    def __init__(self, dist_type, **kwargs):
        super().__init__({"dist_type": dist_type})

    @property
    def dist_type(self):
        """Return the type of the distribution.

        Returns
        -------
        str
            The type of the distribution.
        """
        return self["dist_type"]

    @abstractproperty
    def dist(self):
        """Return the actual distribution.

        Returns
        -------
        chaospy.Distribution
            The actual distribution.
        """

    @abstractproperty
    def uniform_dist(self):
        """Return a uniform distribution used for sampling.

        Returns
        -------
        chaospy.Uniform
            The uniform distribution.
        """

    @abstractproperty
    def salib_name(self):
        """Return the name of the distribution in SALib.

        Returns
        -------
        str
            The name of the distribution in SALib.
        """

    @abstractproperty
    def salib_bounds(self):
        """Return the bounds of the distribution in SALib.

        Returns
        -------
        list [float]
            The bounds of the distribution in SALib.
        """

    @property
    def expect(self):
        """Return the expected value of the distribution.

        Returns
        -------
        float
            The expected value of the distribution.
        """
        return float(chaos.E(self.dist))

    @staticmethod
    def load(dist_type, **kwargs):
        """Load a distribution from its dict representation.

        Returns
        -------
        DistABC
            The loaded distribution.
        """
        from .constant import DistConstant
        from .normal import DistNormal, DistTruncNormal
        from .uniform import DistUniform

        dist_types = {
            "constant": DistConstant,
            "normal": DistNormal,
            "trunc_normal": DistTruncNormal,
            "uniform": DistUniform,
        }
        return dist_types[dist_type](**kwargs)



"""API for `shamo.eeg`."""
from .leadfield.single.problem import ProbEEGLeadfield
from .leadfield.single.solution import SolEEGLeadfield

from .leadfield.parametric.problem import ProbParamEEGLeadfield
from .leadfield.parametric.solution import SolParamEEGLeadfield
from .leadfield.parametric.surrogate import SurrEEGLeadfield, SurrEEGLeadfieldToRef

__all__ = [
    "ProbEEGLeadfield",
    "SolEEGLeadfield",
    "ProbParamEEGLeadfield",
    "SolParamEEGLeadfield",
    "SurrEEGLeadfield",
    "SurrEEGLeadfieldToRef",
]



"""Implement `SolParamABC` class."""
from abc import abstractproperty, abstractmethod
from pathlib import Path
import re

from shamo.core.objects import ObjDir


class SolParamABC(ObjDir):
    """Store information about the solution of a parametric problem.

    Parameters
    ----------
    name : str
        The name of the solution.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the solution.

    Other Parameters
    ----------------
    sub_json_paths : list [str]
        The relative paths to the sub-solutions.
    """

    def __init__(self, name, parent_path, **kwargs):
        super().__init__(name, parent_path)
        self.update(
            {"sub_json_paths": [str(Path(p)) for p in kwargs.get("sub_json_paths", [])]}
        )
        self.sub_sols = self.get_sub_sols()

    @abstractproperty
    def sub_class(self):
        """Return the class of the sub-solutions.

        Returns
        -------
        shamo.core.objects.ObjDir
            The class of the sub-solutions.
        """

    @property
    def sub_json_paths(self):
        """Return the paths to the sub-solutions JSON files.

        Returns
        -------
        pathlib.Path
            The paths to the sub-solutions JSON files.
        """
        return [self.path / p for p in self["sub_json_paths"]]

    @abstractmethod
    def finalize(self, **kwargs):
        """Finalize the solution."""

    def _get_sub_json_paths(self):
        """Get the JSON paths of the sub-solutions."""
        self["sub_json_paths"] = self.get_sub_file(".json")

    def get_sub_sols(self):
        """Return the sub-solutions.

        Returns
        -------
        list [shamo.core.objects.ObjDir]
            The sub-solutions.
        """
        return [self.sub_class.load(p) for p in self.sub_json_paths]

    def get_sub_file(self, suffix):
        """Return the relative paths to the same file in all the sub-solutions.

        Returns
        -------
        list [str]
            The relative paths to the same file in all the sub-solutions.
        """
        return sorted(
            [
                str(p.relative_to(self.path) / f"{p.name}{suffix}")
                for p in self.path.iterdir()
                if p.is_dir and re.match(r"^.+_\d{8}$", p.name)
            ]
        )



"""Implement `CompABC` class."""
from abc import ABC, abstractmethod


class CompABC(dict, ABC):
    """A base class for any problem component.

    Components are the building bricks of all the problems.
    """

    def __init__(self, **kwargs):
        self.update(kwargs)

    @abstractmethod
    def check(self, name, **kwargs):
        """Check if the component is properly configured."""

    @abstractmethod
    def to_pro_param(self, **kwargs):
        """Return an object to be used as a parameter in the rendering of a PRO file."""

    @abstractmethod
    def to_py_param(self, **kwargs):
        """Return an object to be used as a parameter in the rendering of a PY file."""



"""Implement `CompFilePath` class."""
import logging
from pathlib import Path

from .abc import CompABC

logger = logging.getLogger(__name__)


class CompFilePath(CompABC):
    """Store information about a path.

    Other Parameters
    ----------------
    path : pathlib.Path
        The path to the file.
    """

    def __init__(self, **kwargs):
        super().__init__(path=kwargs.get("path", None))

    @property
    def use_path(self):
        """Return ``True`` if a path is set.

        Returns
        -------
        bool
            ``True`` if a path is set, ``False`` otherwise.
        """
        return self["path"] is not None

    @property
    def path(self):
        """Return the path.

        Returns
        -------
        pathlib.Path
            The path.
        """
        return self["path"]

    def set(self, path):
        """Set the path.

        Parameters
        ----------
        path : str, byte or os.PathLike
            The path.
        """
        self["path"] = Path(path)

    def check(self, name, **kwargs):
        """Check if the path is properly set."""
        logger.info(f"Checking file path for '{name}'.")
        if self.use_path:
            if not self["path"].exists():
                raise FileNotFoundError("Argument 'path' does not exist.")

    def to_pro_param(self, **kwargs):
        """Return ``None``."""
        return None

    def to_py_param(self, **kwargs):
        """Return the path."""
        return str(self.path)



"""Implement `ProbHDTDCSSim` class."""
from pathlib import Path
from tempfile import TemporaryDirectory

import nibabel as nib

from shamo.core.problems.single import ProbGetDP, CompSensors, CompGridSampler
from shamo.utils.onelab import pos_to_nii
from .solution import SolHDTDCSSim


class ProbHDTDCSSim(ProbGetDP):
    """A problem definition to simulate HD-tDCS.

    Attributes
    ----------
    sigmas : shamo.core.problems.single.CompTissueProp
        The electrical conductivity of the tissues.
    references : shamo.core.problems.single.CompSensors
        The electrodes to use as the references.
    source : shamo.core.problems.single.CompSensors
        The electrode to use as the source.
    current : float
        The injected current.
    grid : shamo.core.problems.single.CompGridSampler
        The grid sampler if the source space must be based on a grid.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.grid = CompGridSampler()
        self.references = CompSensors()
        self.source = CompSensors()
        self.current = 1.0

    @property
    def template(self):
        """Return the name of the template PRO file.

        Returns
        -------
        str
            The name of the template PRO file.
        """
        return "hdtdcs_simulation.tmplt"

    def solve(self, name, parent_path, model, **kwargs):
        """Simulate HD-tDCS.

        Parameters
        ----------
        name : str
            The name of the solution.
        parent_path : str, byte or os.PathLike
            The path to the parent directory of the solution.
        model : shamo.fem.FEM
            The finite element model to solve the problem for.

        Returns
        -------
        shamo.hd_tdcs.SolHDTDCSSim
            The simulation results.
        """
        for n, t in model.tissues.items():
            self._vol.set(n, t.vol.group)
        shape = [int(s / 10) for s in model.shape]
        if not self.grid.use_grid:
            img = nib.load(model.nii_path)
            mask = img.get_fdata() != 0
            self.grid.set(model.affine, model.shape, mask=mask, resize=False)

        with TemporaryDirectory() as d:
            self._check_components(**model)
            self._gen_pro_file(d, **kwargs, **model)
            self._run_getdp(model, d)
            sol = SolHDTDCSSim(
                name,
                parent_path,
                source=self.source["sensors"],
                references=self.references["sensors"],
                sigmas=self.sigmas,
                current=self.current,
            )
            sol["model_json_path"] = str(sol.get_relative_path(model.json_path))
            for p in Path(d).iterdir():
                if p.suffix == ".pos":
                    self.grid.nii_from_pos(p, sol.path / f"{name}_{p.stem}.nii")
                    p.rename(sol.path / f"{name}_{p.name}")
            sol.save()
        return sol

    def _check_components(self, **kwargs):
        """Check if the components are properly set."""
        super()._check_components(**kwargs)
        self.grid.check("grid", **kwargs)
        self.source.check("source", **kwargs)
        self.references.check("references", **kwargs)

    def _prepare_pro_file_params(self, **kwargs):
        """Return the parameters required to generate the PRO file.

        Returns
        -------
        dict [str, ...]
            The parameters required to generate the PRO file.
        """
        params = super()._prepare_pro_file_params(**kwargs)
        params.update(
            {
                "use_grid": self.grid.to_pro_param(**kwargs),
                "sources": self.source.to_pro_param(**kwargs),
                "sinks": self.references.to_pro_param(**kwargs),
                "current": self.current,
            }
        )
        return params

    def _prepare_py_file_params(self, **kwargs):
        """Return the parameters required to generate the PY file.

        Returns
        -------
        dict [str, ...]
            The parameters required to generate the PY file.
        """
        params = super()._prepare_py_file_params(**kwargs)
        params.update(
            {
                "source": self.source.to_py_param(**kwargs),
                "references": self.references.to_py_param(**kwargs),
                "current": self.current,
                "use_grid": self.grid.use_grid,
            }
        )
        if self.grid.use_grid:
            params["grid"] = self.grid.to_py_param(**kwargs)
        return params



"""Implement the `ObjFile` class."""
import json

from .abc import ObjABC


class ObjFile(ObjABC):
    """A base class for any savable/loadable object stored as a single JSON file.

    Parameters
    ----------
    name : str
        The name of the object.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the object.

    See Also
    --------
    shamo.core.ObjABC
    """

    @property
    def path(self):
        """An alias for `json_path`.

        Returns
        -------
        pathlib.Path
            The path to the object JSON file.
        """
        return self.json_path

    @property
    def json_path(self):
        """Return the path to the object JSON file.

        Returns
        -------
        pathlib.Path
            The path to the object JSON file.
        """
        return self.parent_path / f"{self.name}.json"

    @classmethod
    def _split_json_path(cls, json_path):
        """Return the path and the name of the object from the path of the JSON file.

        Parameters
        ----------
        json_path : pathlib.Path
            The path to the JSON file containing the object data.

        Returns
        -------
        str
            The name of the object.
        pathlib.Path
            The path to the parent directory of the object.
        """
        return json_path.stem, json_path.parent



"""Implement the `ObjABC` class."""
from abc import ABC, abstractclassmethod, abstractproperty
import json
import logging
from pathlib import Path

logger = logging.getLogger(__name__)


class ObjABC(dict, ABC):
    """A base class for any savable/loadable object.

    All the main object can be saved to/loaded from human readable JSON files to make it
    easily compatible with BIDS standars and simplify interactions with other scripts.

    Parameters
    ----------
    name : str
        The name of the object.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the object.

    Raises
    ------
    TypeError
        If argument `name` is not a `str`.
        If argument `parent_path` is not a `str`, `byte` or `os.PathLike`.
    """

    def __init__(self, name, parent_path):
        super().__init__()
        if not isinstance(name, str):
            raise TypeError(f"Argument 'name' expects type str, not {type(name)}.")
        self._name = name
        self._parent_path = Path(parent_path)

    @property
    def name(self):
        """Return the name of the object.

        Returns
        -------
        str
            The name of the object.
        """
        return self._name

    @property
    def parent_path(self):
        """Return the path to the parent directory of the object.

        Returns
        -------
        pathlib.Path
            The path to the parent directory of the object.
        """
        return self._parent_path

    @abstractproperty
    def path(self):
        """Return the path to the object.

        Returns
        -------
        pathlib.Path
            The path to the object.
        """

    @abstractproperty
    def json_path(self):
        """Return the path to the object JSON file.

        Returns
        -------
        pathlib.Path
            The path to the object JSON file.
        """
        pass

    def save(self, exist_ok=True):
        """Save the object to a JSON file.

        Parameters
        ----------
        exist_ok : bool, optional
            If set to ``True``, any already existing object is overriden. Otherwise, if
            the object already exists, a `FileExistsError`. The default is ``True``.

        Raises
        ------
        FileExistsError
            If `exist_ok` is set to ``False`` and the object already exists.
        TypeError
            If any of the keys/values to be stored is not a `str`, `int`, `float`,
            `bool` or ``None``.
        """
        self.parent_path.mkdir(parents=True, exist_ok=True)
        if not exist_ok and self.json_path.exists():
            raise FileExistsError(
                (
                    f"File '{str(self.json_path)}' already exists. "
                    "If you want to override it, set argument 'exist_ok' to True."
                )
            )
        with open(self.json_path, "w") as f:
            json.dump(self, f, indent=4, sort_keys=True)

    @abstractclassmethod
    def _split_json_path(cls, json_path):
        """Return the path and the name of the object from the path of the JSON file.

        Parameters
        ----------
        json_path : pathlib.Path
            The path to the JSON file containing the object data.

        Returns
        -------
        str
            The name of the object.
        pathlib.Path
            The path to the parent directory of the object.
        """
        pass

    @classmethod
    def load(cls, json_path):
        """Load an object from a JSON file.

        Parameters
        ----------
        json_path : str, byte or os.PathLike
            The path to the JSON file containing the object data.

        Raises
        ------
        TypeError
            If argument `json_path` is not a `str`, `byte` or `os.PathLike`.

        """
        json_path = Path(json_path)
        if not json_path.exists():
            raise FileNotFoundError(f"File '{str(json_path)}' does not exist.")
        if not json_path.suffix == ".json":
            raise ValueError(f"Argument 'json_path' must end with '.json'.")
        with open(json_path, "r") as f:
            data = json.load(f)
        return cls(*cls._split_json_path(json_path), **data)



"""Implement `ProbParamEEGLeadfield` class."""
import logging

from shamo.core.problems.parametric import ProbParamGetDP
from shamo.core.problems.single import (
    CompSensors,
    CompTissues,
    CompGridSampler,
    CompFilePath,
)
from shamo.eeg import ProbEEGLeadfield
from .solution import SolParamEEGLeadfield
from shamo.utils.path import get_relative_path

logger = logging.getLogger(__name__)


class ProbParamEEGLeadfield(ProbParamGetDP):
    """A problem definition to solve the parametric EEG forward problem.

    Attributes
    ----------
    sigmas : shamo.core.problems.parametric.CompParamTissueProp
        The electrical conductivity of the tissues.
    markers : shamo.core.problems.single.CompSensors
        The electrodes to ignore when solving the problem.
    reference : shamo.core.problems.single.CompSensors
        The electrode to use as the reference.
    rois : shamo.core.problems.single.CompTissues
        The tissues in which the leadfield must be computed.
    elems_path : shamo.core.problems.single.CompFilePath
        The path to the elements subset.
    grid : shamo.core.problems.single.CompGridSampler
        The grid sampler if the source space must be based on a grid.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._source = CompSensors()
        self.markers = CompSensors()
        self.reference = CompSensors()
        self.rois = CompTissues()
        self.elems_path = CompFilePath()
        self.grid = CompGridSampler()

    @property
    def template(self):
        """Return the name of the template PY file.

        Returns
        -------
        str
            The name of the template PY file.
        """
        return "eeg_leadfield.tmplt"

    def solve(
        self,
        name,
        parent_path,
        model,
        n_evals,
        method="sequential",
        n_proc=1,
        skip=0,
        **kwargs,
    ):
        """Solve the EEG parametric forward problem.

        Parameters
        ----------
        name : str
            The name of the solution.
        parent_path : str, byte or os.PathLike
            The path to the parent directory of the solution.
        model : shamo.fem.FEM
            The finite element model to solve the problem for.
        n_evals : int
            The number of evaluation points.
        method : str, optional
            The method to solve the sub-problems. The accepted values are
            ``'sequential'``, ``'multiprocessing'``, ``'job'``. (The default is
            ``'sequential'``)
        n_proc : int, optional
            The number of processes to solve the problem when `method` is set to
            ``'multiprocessing'``. (The default is ``1``)
        skip : int, optional
            The number of points to skip at the beginning of the sequence. (The default
            is ``0``)

        Returns
        -------
        shamo.eeg.SolParamEEGLeadfield
            The parametric EEG leafield object.
        """
        logger.info("Solving problem")
        self._check_components(**model, **kwargs)
        params = self._gen_params(n_evals, skip, **kwargs)
        base_path = f"{parent_path}/{name}" if method == "jobs" else "."
        sub_probs = self._gen_sub_probs(
            n_evals, parent_path=base_path, **params, **kwargs
        )
        sol = SolParamEEGLeadfield(
            name,
            parent_path,
            markers=self.markers["sensors"],
            reference=self.reference["sensors"],
            rois=self.rois["tissues"],
            sigmas=self.sigmas,
            use_grid=self.grid.use_grid,
        )
        sol["model_json_path"] = str(sol.get_relative_path(model.json_path))
        sub_sols = self._solve_sub_probs(
            sol, sub_probs, method=method, n_proc=n_proc, model=model, **kwargs
        )
        return sol

    def _gen_fixed_varying(self, **kwargs):
        """Generate two lists containing the fixed and varying parameters.

        Returns
        -------
        list [dict {str, ...}]
            The fixed parameters.
        list [dict {str, ...}]
            The varying parameters.
        """
        return self.sigmas.to_param("sigmas", **kwargs)

    def _gen_sub_prob(self, i, **kwargs):
        """Generate one sub-problem.

        Parameters
        ----------
        i : int
            The index of the sub-problem.

        Returns
        -------
        shamo.eeg.ProbEEGLeadfield
            The generated sub-problem.

        Other Parameters
        ----------------
        sigmas : dict [str, list[numpy.ndarray, str]]
            The electrical conductivity of the tissues.
        """
        s = kwargs.get("sigmas", {})
        prob = ProbEEGLeadfield()
        for t, p in s.items():
            prob.sigmas.set(t, p[0][i], p[1])
        prob.reference = self.reference
        prob.markers = self.markers
        prob.rois = self.rois
        if self.elems_path.use_path:
            prob.elems_path.set(
                get_relative_path(kwargs.get("parent_path", "."), self.elems_path.path)
            )
        prob.grid = self.grid
        return prob

    def _check_components(self, **kwargs):
        """Check if the components are properly set."""
        self.sigmas.check("sigmas", **kwargs)
        self._source.check("source", **kwargs)
        self.reference.check("reference", **kwargs)
        self.rois.check("region of interest", **kwargs)
        self.elems_path.check("elements path", **kwargs)
        self.grid.check("grid", **kwargs)
        if self.elems_path.use_path and self.grid.use_grid:
            raise RuntimeError(
                "Both 'elems_path' and 'grid' are set. Only one of them is allowed."
            )



"""API for `shamo.utils.path`."""
from pathlib import Path


def get_relative_path(ref, path):
    """Return the relative path from the reference to a file or directory.

    Parameters
    ----------
    ref : str, byte or os.PathLike
        The path to the reference file or directory.
    path : str, byte or os.PathLike
        The path to the file or directory to compute the relative path for.

    Returns
    -------
    pathlib.Path
        The relative path to the file or directory.
    """
    ref = Path(ref).expanduser().resolve()
    path = Path(path).expanduser().resolve()
    here = False
    rel_path = ""
    while not here:
        try:
            current = str(path.relative_to(ref))
            return Path(rel_path + current)
        except ValueError:
            rel_path += "../"
            ref = ref.parent



"""API for `shamo.core.problems.parametric`."""
from .abc import ProbParamABC
from .getdp import ProbParamGetDP

from .components.tissue_property import CompParamTissueProp
from .components.value import CompParamValue


__all__ = ["ProbParamABC", "ProbParamGetDP", "CompParamTissueProp", "CompParamValue"]



"""Implement the `FEM` class."""
from collections.abc import Mapping, Iterable
from functools import partialmethod
import logging
from pathlib import Path
from pprint import pformat
from tempfile import TemporaryDirectory

import gmsh
import meshio
import nibabel as nib
from nilearn.image import crop_img
import numpy as np
import pygalmesh as cgal
from scipy.interpolate import RegularGridInterpolator
from scipy.spatial.distance import cdist

from shamo.core.objects import ObjDir
from . import Field, Group, Tissue, Sensor, PointSensor

logger = logging.getLogger(__name__)


class FEM(ObjDir):
    """A finite element model.

    Parameters
    ----------
    name : str
        The name of the model.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the model.
    """

    def __init__(self, name, parent_path, **kwargs):
        super().__init__(name, parent_path)
        self.update(
            {
                "tissues": {
                    t: Tissue(**d) for t, d in kwargs.get("tissues", {}).items()
                },
                "mesh_params": kwargs.get("mesh_params", {}),
                "sensors": {
                    s: self._load_sensor(d)
                    for s, d in kwargs.get("sensors", {}).items()
                },
            }
        )
        logger.info(f"Model '{name}' initialized in '{parent_path}'  directory.")

    @property
    def nii_path(self):
        """Return the path to the NIFTI file.

        Returns
        -------
        pathlib.Path
            The path to the NIFTI file.
        """
        return self.path / f"{self.name}.nii"

    @property
    def affine(self):
        """Return the affine matrix of the NIFTI file.

        Returns
        -------
        numpy.ndarray
            The affine matrix of the NIFTI file.
        """
        img = nib.load(self.nii_path)
        return img.affine

    @property
    def shape(self):
        """Return the shape of the NIFTI file.

        Returns
        -------
        numpy.ndarray
            The shape of the NIFTI file.
        """
        img = nib.load(self.nii_path)
        return img.shape

    @property
    def mesh_path(self):
        """Return the path to the MSH file.

        Returns
        -------
        pathlib.Path
            The path to the MSH file.
        """
        return self.path / f"{self.name}.msh"

    @property
    def tissues(self):
        """Return the tissues of the model.

        Returns
        -------
        dict [str, shamo.fem.Tissue]
            The tissues of the model.
        """
        return self["tissues"]

    @property
    def sensors(self):
        """Return the sensors of the model.

        Returns
        -------
        dict [str, shamo.fem.Sensor]
            The sensors of the model.
        """
        return self["sensors"]

    @property
    def mesh_params(self):
        """Return the parameters used to produce the mesh.

        Returns
        -------
        dict [str, float]
            The parameters used to produce the mesh.
        """
        return self["mesh_params"]

    # Mesh -----------------------------------------------------------------------------

    def mesh_from_array(self, labels, affine, tissues, **kwargs):
        """Generate a MSH file from a labelled array.

        Parameters
        ----------
        labels : numpy.ndarray
            A labelled array corresponding to a multi-segments image. The array must
            contain int labels starting from ``0`` (air) without skipping a number.
        affine : numpy.ndarray
            The affine matrix of the volume.
        tissues : iterable [str]
            The names of the tissues in the same order as the labels.

        Other Parameters
        ----------------
        lloyd : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        odt : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        perturb : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        exude : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_edge_size_at_feature_edges: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        min_facet_angle: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_radius_surface_delaunay_ball: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_cell_circumradius: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_facet_distance: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_circumradius_edge_ratio: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        seed : int, optional
            For more information, refer to the documentation of `pygalmesh`.

        Raises
        ------
        TypeError
            If argument 'labels' is not a `numpy.ndarray`.
            If argument 'affine' is not a `numpy.ndarray`.
            If an element of argument 'tissues' is not a `str`.
        ValueError
            If argument 'labels' is not a 3D array.
            If argument 'affine' is neither a (3, 4) nor a (4, 4) array.
            If argument 'tissues' does not contain as many elements as there are labels
            in 'labels'.

        See Also
        --------
        pygalmesh.generate_from_array
        """
        if not isinstance(labels, np.ndarray):
            raise TypeError("Argument 'labels' expects type numpy.ndarray.")
        labels = labels.astype(np.uint8)
        if labels.ndim != 3:
            raise ValueError("Argument 'labels' must be a 3D array.")
        if not isinstance(affine, np.ndarray):
            raise TypeError("Argument 'affine' expects type numpy.ndarray.")
        if affine.shape not in ((3, 4), (4, 4)):
            raise ValueError("Argument 'affine' expects shape (3,4) or (4,4).")
        affine = np.vstack((affine[:3, :], np.array([0, 0, 0, 1])))
        # Convert [mm] to [m]
        affine = np.diag([1e-3] * 3 + [1]) @ affine
        tissues = list(tissues)
        for t in tissues:
            if not isinstance(t, str):
                raise TypeError("Argument 'tissues' expects an iterable of str.")
        if len(tissues) != np.max(labels):
            raise ValueError(
                (
                    "Argument 'tissues' must contain as many names "
                    "as there are unique labels in 'labels'."
                )
            )

        img = nib.Nifti1Image(labels, affine)
        cropped_img = crop_img(img)
        cropped_img.to_filename(self.nii_path)
        labels = cropped_img.get_fdata().astype(np.uint16)
        affine = cropped_img.affine

        with TemporaryDirectory() as d:
            self._gen_init_mesh(labels, np.eye(4), d, **kwargs)
            self._apply_transform(affine, d)
            self._add_tissues(tissues, d)
        self.save()
        logger.info("Mesh generated.")

    def mesh_from_nii(self, nii_path, tissues, **kwargs):
        """Generate a MSH file from a labelled NIFTI file.

        Parameters
        ----------
        nii_path : str, byte or os.PathLike
            The path to the NIFTI file containing the labelled volume.
        tissues : iterable [str]
            The names of the tissues in the same order as the labels.

        Other Parameters
        ----------------
        lloyd : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        odt : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        perturb : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        exude : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_edge_size_at_feature_edges: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        min_facet_angle: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_radius_surface_delaunay_ball: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_cell_circumradius: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_facet_distance: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_circumradius_edge_ratio: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        seed : int, optional
            For more information, refer to the documentation of `pygalmesh`.

        Raises
        ------
        TypeError
            If argument `nii_path` is not a `str`, `byte` or `os.PathLike`.

        See Also
        --------
        FEM.mesh_from_array
        pygalmesh.generate_from_array
        """
        nii_path = Path(nii_path)

        img = nib.load(str(nii_path))
        return self.mesh_from_array(
            img.get_fdata().astype(np.uint8), img.affine, tissues, **kwargs
        )

    def mesh_from_masks(self, masks, affine, **kwargs):
        """Generate a MSH file from multiple binary masks.

        Parameters
        ----------
        mask : Mapping [str, numpy.ndarray]
            A mapping from the names of the tissues to the corresponding binary masks.
        affine : numpy.ndarray
            The affine matrix of the volume.

        Other Parameters
        ----------------
        lloyd : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        odt : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        perturb : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        exude : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_edge_size_at_feature_edges: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        min_facet_angle: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_radius_surface_delaunay_ball: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_cell_circumradius: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_facet_distance: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_circumradius_edge_ratio: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        seed : int, optional
            For more information, refer to the documentation of `pygalmesh`.

        Raises
        ------
        TypeError
            If argument `masks` is not a mapping from `str` to `numpy.ndarray`.
        ValueError
            If the masks in 'masks' are not all of the same shape.

        See Also
        --------
        FEM.mesh_from_array
        pygalmesh.generate_from_array
        """
        if not isinstance(masks, Mapping):
            raise TypeError(
                "Argument 'masks' expects a mapping from str to numpy.ndarray."
            )
        for t in masks.keys():
            if not isinstance(t, str):
                raise TypeError(
                    "Argument 'masks' expects a mapping from str to numpy.ndarray."
                )
        shape = None
        for m in masks.values():
            if not isinstance(m, np.ndarray):
                raise TypeError(
                    "Argument 'masks' expects a mapping from str to numpy.ndarray."
                )
            if shape is not None and m.shape != shape:
                raise ValueError(
                    "Values in argument 'masks' must all have the same shape."
                )
            shape = m.shape

        labels = np.zeros(list(masks.values())[0].shape)
        tissues = []
        for l, (t, m) in enumerate(masks.items()):
            labels[m.astype(bool)] = l + 1
            tissues.append(t)
        return self.mesh_from_array(labels, affine, tissues, **kwargs)

    def mesh_from_niis(self, niis, **kwargs):
        """Generate a MSH file from multiple binary masks.

        Parameters
        ----------
        niis : Mapping [str, str, byte or os.PathLike]
            A mapping from the names of the tissues to the path to the corresponding
            NIFTI binary masks.

        Other Parameters
        ----------------
        lloyd : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        odt : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        perturb : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        exude : bool, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_edge_size_at_feature_edges: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        min_facet_angle: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_radius_surface_delaunay_ball: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_cell_circumradius: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_facet_distance: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        max_circumradius_edge_ratio: float, optional
            For more information, refer to the documentation of `pygalmesh`.
        seed : int, optional
            For more information, refer to the documentation of `pygalmesh`.

        Raises
        ------
        TypeError
            If argument `niis` is not a mapping from `str` to `str`, `byte` or
            `os.PathLike`.

        See Also
        --------
        FEM.mesh_from_masks
        FEM.mesh_from_array
        pygalmesh.generate_from_array
        """
        if not isinstance(niis, Mapping):
            raise TypeError(
                (
                    "Argument 'niis' expects a mapping from str "
                    "to str, byte or os.PathLike."
                )
            )
        for t, p in niis.items():
            niis[t] = Path(p)

        imgs = {t: nib.load(str(p)) for t, p in niis.items()}
        return self.mesh_from_masks(
            {t: i.get_fdata() for t, i in imgs.items()},
            list(imgs.values())[0].affine,
            **kwargs,
        )

    def _gen_init_mesh(self, labels, affine, tmp_dir, **kwargs):
        """Generate the initial mesh usign CGAL."""
        init_mesh = cgal.generate_from_array(
            labels, nib.affines.voxel_sizes(affine), **kwargs
        )
        meshio.write(str(Path(tmp_dir) / "init_mesh.mesh"), init_mesh)
        self["mesh_params"] = kwargs
        logger.info("Initial mesh generated.")

    def _apply_transform(self, affine, tmp_dir):
        """Apply the affine transform to the mesh."""
        gmsh.initialize()
        gmsh.option.setNumber("General.Terminal", 1)
        gmsh.merge(str(Path(tmp_dir) / "init_mesh.mesh"))
        gmsh.plugin.run("NewView")
        axes = ["x", "y", "z"]
        for r in range(3):
            for c in range(3):
                gmsh.plugin.setNumber(
                    "Transform", "A{}{}".format(r + 1, c + 1), affine[r, c]
                )
            gmsh.plugin.setNumber("Transform", "T{}".format(axes[r]), affine[r, -1])
        gmsh.plugin.run("Transform")
        # gmsh.model.mesh.reclassifyNodes()
        gmsh.option.setNumber("Mesh.Binary", 1)
        gmsh.write(str(Path(tmp_dir) / "init_mesh.msh"))
        gmsh.finalize()
        logger.info("Affine transformation applied.")

    def _add_tissues(self, tissues, tmp_dir):
        """Add the tissues as physical groups."""
        gmsh.initialize()
        gmsh.option.setNumber("General.Terminal", 1)
        gmsh.merge(str(Path(tmp_dir) / "init_mesh.msh"))
        for l, t in enumerate(tissues):
            entity = l + 1
            surf_group = gmsh.model.addPhysicalGroup(2, [entity])
            gmsh.model.setPhysicalName(2, surf_group, t)
            vol_group = gmsh.model.addPhysicalGroup(3, [entity])
            gmsh.model.setPhysicalName(3, vol_group, t)
            self["tissues"][t] = Tissue(
                Group(2, [entity], surf_group), Group(3, [entity], vol_group)
            )
            logger.info(f"Tissue '{t}' added.")
        gmsh.option.setNumber("Mesh.Binary", 1)
        gmsh.write(str(self.mesh_path))
        gmsh.finalize()

    # Sensors --------------------------------------------------------------------------

    def add_point_sensor(self, name, coords, tissue, dim):
        """Add a point sensor to the mesh.

        Parameters
        ----------
        name : str
            The name of the sensor.
        coords : Iterable [float] or numpy.ndarray
            The coordinates of the sensor.
        tissue : str
            The name of the tissue the sensor is on/in.
        dim : int
            If set to ``2``, the sensor is added on the surface of the tissue. If set to
            ``3``, it is placed inside the tissue.

        Raises
        ------
        TypeError
            If argument `coords` is neither an `Iterable` or a `numpy.ndarray`.
            If argument `dim` is not an `int`.
        ValueError
            If argument `name` refers to an existing sensor.
            If argument `coords` is not a proper 3D coordinate.
            If argument `tissue` refers to a non existing tissue.
        """
        if name in self.sensors:
            raise ValueError(f"Sensor '{name}' already exists in model.")
        if not isinstance(coords, (Iterable, np.ndarray)):
            raise TypeError("Argument 'coords' expects type Iterable or numpy.ndarray.")
        if len(coords) != 3:
            raise ValueError("Argument 'coords' must be a 3D coordinate.")
        coords = tuple([1e-3 * c for c in coords])
        if tissue not in self.tissues:
            raise ValueError(f"Tissue '{tissue}' not found in model.")
        dim = int(dim)

        logger.debug(
            (
                f"Adding 1 sensor {'on' if dim == 2 else 'in'} "
                f"tissue '{tissue}' with coords:\n{coords}"
            )
        )
        gmsh.initialize()
        gmsh.option.setNumber("General.Terminal", 1)
        gmsh.open(str(self.mesh_path))
        nodes_tags, nodes_coords = self._get_tissue_nodes(tissue, dim)
        self._add_point_sensor(name, coords, nodes_tags, nodes_coords, tissue)
        gmsh.model.mesh.removeDuplicateNodes()
        gmsh.option.setNumber("Mesh.Binary", 1)
        gmsh.write(str(self.mesh_path))
        gmsh.finalize()
        self.save()
        logger.info(f"Sensor '{name}' added.")

    add_point_sensor_on = partialmethod(add_point_sensor, dim=2)
    add_point_sensor_in = partialmethod(add_point_sensor, dim=3)

    def add_point_sensors(self, coords, tissue, dim):
        """Add multiple point sensors to the mesh.

        Parameters
        ----------
        coords : Mapping [str, Iterable [float]]
            The coordinates of the sensor.
        tissue : str
            The name of the tissue the sensor is on/in.
        dim : int
            If set to ``2``, the sensor is added on the surface of the tissue. If set to
            ``3``, it is placed inside the tissue.

        Raises
        ------
        TypeError
            If argument `coords` is not a Mapping of coordinates.
            If argument `dim` is not an `int`.
        ValueError
            If argument `tissue` refers to a non existing tissue.
        """
        coords = {s: tuple([1e-3 * x for x in c]) for s, c in coords.items()}
        if tissue not in self.tissues:
            raise ValueError(f"Tissue '{tissue}' not found in model.")
        dim = int(dim)

        logger.debug(
            (
                f"Adding {len(coords)} sensors {'on' if dim == 2 else 'in'} "
                f"tissue '{tissue}' with coords:\n{pformat(coords)}"
            )
        )
        gmsh.initialize()
        gmsh.option.setNumber("General.Terminal", 1)
        gmsh.open(str(self.mesh_path))
        nodes_tags, nodes_coords = self._get_tissue_nodes(tissue, dim)
        for s, c in coords.items():
            self._add_point_sensor(s, c, nodes_tags, nodes_coords, tissue)
            logger.info(
                f"Sensor '{s}' added {'on' if dim == 2 else 'in'} tissue '{tissue}'."
            )
        gmsh.model.mesh.removeDuplicateNodes()
        gmsh.option.setNumber("Mesh.Binary", 1)
        gmsh.write(str(self.mesh_path))
        gmsh.finalize()
        self.save()

    add_point_sensors_on = partialmethod(add_point_sensors, dim=2)
    add_point_sensors_in = partialmethod(add_point_sensors, dim=3)

    def add_point_sensors_from_tsv(self, tsv_path, tissue, dim):
        """Add multiple point sensors to the mesh from a TSV file.

        Parameters
        ----------
        tsv_path : str, byte or os.PathLike
            The path to the TSV file.
        tissue : str
            The name of the tissue the sensor is on/in.
        dim : int
            If set to ``2``, the sensor is added on the surface of the tissue. If set to
            ``3``, it is placed inside the tissue.

        Raises
        ------
        TypeError
            If argument `coords` is not a Mapping of coordinates.
            If argument `dim` is not an `int`.
        ValueError
            If argument `tissue` refers to a non existing tissue.
        """
        tsv_path = Path(tsv_path)
        data = np.genfromtxt(
            tsv_path, delimiter="\t", skip_header=1, dtype=None, encoding="utf-8"
        )
        coords = {d[0]: [d[1], d[2], d[3]] for d in data}
        logger.info(f"{len(coords)} sensors coordinates extracted from '{tsv_path}'.")
        logger.debug(pformat(coords))
        return self.add_point_sensors(coords, tissue, dim)

    add_point_sensors_from_tsv_on = partialmethod(add_point_sensors_from_tsv, dim=2)
    add_point_sensors_from_tsv_in = partialmethod(add_point_sensors_from_tsv, dim=3)

    def _get_tissue_nodes(self, tissue, dim):
        """Return all the nodes of the tissue in specified dimensio entities."""
        if dim == 2:
            entities = self.tissues[tissue].surf.entities
        elif dim == 3:
            entities = self.tissues[tissue].vol.entities
        logger.debug(
            (
                f"Acquiring nodes from tissue '{tissue}' "
                f"{'surface' if dim == 2 else 'volume'} with entities:\n{entities}"
            )
        )
        logger.debug(
            f"All {'surface' if dim == 2 else 'volume'} entities:"
            f"\n{gmsh.model.getEntities(dim)}"
        )
        nodes = [gmsh.model.mesh.getNodes(dim, e, True)[:2] for e in entities]
        logger.debug(f"Acquired {nodes[0][0].size} nodes:\n{nodes}")
        tags = np.hstack([t for t, _ in nodes])
        coords = np.hstack([c for _, c in nodes]).reshape((-1, 3))
        return tags, coords

    _get_tissue_surf_nodes = partialmethod(_get_tissue_nodes, dim=2)
    _get_tissue_vol_nodes = partialmethod(_get_tissue_nodes, dim=3)

    def _add_point_sensor(self, name, coords, nodes_tags, nodes_coords, tissue):
        """Add a point sensor."""
        dist = cdist([coords], nodes_coords).ravel()
        min_dist_idx = np.argmin(dist)
        node_tag = nodes_tags[min_dist_idx]
        mesh_coords = nodes_coords[min_dist_idx, :].ravel()
        entity, group = self._add_point_sensor_on_node(name, node_tag, mesh_coords)
        sensor = PointSensor(
            tissue, coords, mesh_coords, Group(0, [entity], group), node_tag
        )
        self["sensors"][name] = sensor

    def _add_point_sensor_on_node(self, name, node_tag, node_coords):
        """Add a point sensor on a node."""
        entity = gmsh.model.addDiscreteEntity(0)
        gmsh.model.mesh.addNodes(0, entity, [node_tag], node_coords)
        gmsh.model.mesh.addElementsByType(entity, 15, [], [node_tag])
        max_group = max([tag for dim, tag in gmsh.model.getPhysicalGroups(-1)])
        group = gmsh.model.addPhysicalGroup(0, [entity], max_group + 1)
        gmsh.model.setPhysicalName(0, group, name)
        return entity, group

    def _load_sensor(self, sensor):
        """Load a sensor depending on its type."""
        if sensor["sensor_type"] == Sensor.TYPE_POINT:
            return PointSensor(**sensor)

    # Fields ---------------------------------------------------------------------------

    def field_from_elems(
        self, name, tissue, elems_tags, elems_vals, fill_val=None, formula="1"
    ):
        """Add a field to the mesh from element data.

        Parameters
        ----------
        name : str
            The name of the field. In the mesh, it will be added as '{tissue}_{name}'.
        tissue : str
            The tissue in which the field must be added.
        elems_tags : np.ndarray or Iterable [int]
            The tags of the elements in which the value is available.
        elems_vals : np.ndarray or Iterable
            The values of the field on the corresponding elements.
        fill_val : np.ndarray or Iterable
            The value to add in elements of the tissue that are not in `elems_tags`.
        formula : str
            The formula linking the field to a physical property.

        Raises
        ------
        KeyError
            If argument `tissue` does not correspond to a tissue of the mesh.
            If argument `name` refers to an already existing field.
        TypeError
            If argument `elems_tags` is neither a `numpy.ndarray` or a `Iterable` of
            `int`.
            If argument `elems_vals` is neither a `numpy.ndarray` or a `Iterable`.
            If argument `fill_val` is neither a `numpy.ndarray` or a `Iterable`.
        ValueError
            If argument `elems_vals` does not contain a scalar, vector or tensor field.
            If argument `fill_val` does not correspond to the values in `elems_vals`.
        """
        if tissue not in self.tissues:
            raise KeyError(f"Tissue '{tissue}' not found in model.")
        if name in self.tissues[tissue].fields:
            raise KeyError(f"Field '{name}' already exists in tissue '{tissue}'.")
        if not isinstance(elems_tags, (np.ndarray, Iterable)):
            raise TypeError(
                "Argument 'elems_tags' expects numpy.ndarray or Iterable of int."
            )
        elems_tags = np.array(elems_tags)
        if not isinstance(elems_vals, (np.ndarray, Iterable)):
            raise TypeError(
                "Argument 'elems_vals' expects numpy.ndarray or Iterable of int."
            )
        elems_vals = np.array(elems_vals).ravel()
        n_vals = int(elems_vals.size / elems_tags.size)
        field_types = {1: Field.SCALAR, 3: Field.VECTOR, 9: Field.TENSOR}
        if n_vals not in field_types:
            raise ValueError(
                "Argument 'elems_vals' must contain scalar, vector or tensor values."
            )
        field_type = field_types[n_vals]
        if not isinstance(fill_val, (np.ndarray, Iterable)):
            raise TypeError("Argument 'fill_val' expects numpy.ndarray or Iterable.")
        fill_val = np.array(fill_val).ravel()
        if fill_val.size != n_vals:
            raise ValueError(
                "Argument 'fill_val' must be of the same type and size as 'elems_vals'."
            )
        # TODO: Check formula

        gmsh.initialize()
        gmsh.open(str(self.mesh_path))
        elems_tags, elems_vals = self._fill_empty_field_elems(
            tissue, elems_tags, elems_vals, n_vals, fill_val
        )
        view = self._add_field_view(name, tissue, elems_tags, elems_vals, n_vals)
        gmsh.option.setNumber("PostProcessing.SaveMesh", 0)
        gmsh.option.setNumber("Mesh.Binary", 1)
        gmsh.view.write(view, str(self.mesh_path), True)
        gmsh.finalize()
        self.tissues[tissue].fields[name] = Field(field_type, view, formula)
        self.save()
        logger.info(f"Field '{name}' added in tissue '{tissue}'.")

    def field_from_array(
        self,
        name,
        field,
        affine,
        tissue,
        fill_val,
        formula="1",
        nearest=True,
        resize=True,
    ):
        """Add a field to the mesh from an array.

        Parameters
        ----------
        name : str
            The name of the field. In the mesh, it will be added as '{tissue}_{name}'.
        field : numpy.ndarray
            The field values.
        affine : numpy.ndarray
            The affine matrix of the field.
        tissue : str
            The tissue in which the field must be added.
        fill_val : np.ndarray or Iterable
            The value to add in elements of the tissue that are not in `elems_tags`.
        formula : str, optional
            The formula linking the field to a physical property. The default value is
            `'1'`.
        nearest : bool, optional
            If set to ``True``, no interpolation is performed. Otherwise, a linear
            interpolation is used. The default value is ``True``.
        resize : bool, optional
            If set to ``True``, the affine matrix is converted from [mm] to [m]. The
            default is ``True``.

        Raises
        ------
        TypeError
            If argument `field` is not a `numpy.ndarray`.
            If argument `affine` is not a `numpy.ndarray`.
        ValueError
            If argument `field` is not a 3D or 4D array containing a scalar, vector or
            tensor field.
            If argument `affine` is neither a (3, 4) nor a (4, 4) array.
        KeyError
            If argument `tissue` does not correspond to a tissue of the mesh.

        See Also
        --------
        FEM.field_from_array
        """
        if not isinstance(field, np.ndarray):
            raise TypeError("Argument 'field' expects type numpy.ndarray.")
        if field.ndim not in (3, 4):
            raise ValueError(
                (
                    "Argument 'field' must be a 3D or 4D array containing scalar, "
                    "vector or tensor values."
                )
            )
        if field.ndim == 4 and field.shape[-1] not in (1, 3, 9):
            raise ValueError(
                (
                    "Argument 'field' must be a 3D or 4D array containing scalar, "
                    "vector or tensor values."
                )
            )
        if not isinstance(affine, np.ndarray):
            raise TypeError("Argument 'affine' expects type numpy.ndarray.")
        if affine.shape not in ((3, 4), (4, 4)):
            raise ValueError("Argument 'affine' expects shape (3,4) or (4,4).")
        affine = np.vstack((affine[:3, :], np.array([0, 0, 0, 1])))
        # Convert [mm] to [m]
        if resize:
            affine = np.diag([1e-3] * 3 + [1]) @ affine
        if tissue not in self.tissues:
            raise KeyError(f"Tissue '{tissue}' not found in model.")

        coords = self._get_field_coords(field.shape, affine)
        elems_tags, elems_vals = self._interp_field(
            tissue, coords, field, "nearest" if nearest else "linear", fill_val
        )
        return self.field_from_elems(
            name, tissue, elems_tags, elems_vals, fill_val, formula
        )

    def field_from_nii(
        self, name, nii_path, tissue, fill_val, formula="1", nearest=True, resize=True
    ):
        """Add a field to the mesh from a NIFTI image.

        Parameters
        ----------
        name : str
            The name of the field. In the mesh, it will be added as '{tissue}_{name}'.
        nii_path : str, byte or os.PathLike
            The path to the NIFTI file containing the field.
        tissue : str
            The tissue in which the field must be added.
        fill_val : np.ndarray or Iterable
            The value to add in elements of the tissue that are not in `elems_tags`.
        formula : str, optional
            The formula linking the field to a physical property. The default value is
            `'1'`.
        nearest : bool, optional
            If set to ``True``, no interpolation is performed. Otherwise, a linear
            interpolation is used. The default value is ``True``.
        resize : bool, optional
            If set to ``True``, the affine matrix is converted from [mm] to [m]. The
            default is ``True``.

        Raises
        ------
        TypeError
            If argument `nii_path` is not a `str`, `byte` or `os.PathLike`.

        See Also
        --------
        FEM.field_from_array
        """
        nii_path = Path(nii_path)

        img = nib.load(str(nii_path))
        return self.field_from_array(
            name,
            img.get_fdata(),
            img.affine,
            tissue,
            fill_val,
            formula,
            nearest,
            resize,
        )

    def _get_tissue_elems(self, tissue, dim):
        """Return all the elements of the tissue in specified dimension."""
        if dim == 2:
            entities = self.tissues[tissue].surf.entities
        elif dim == 3:
            entities = self.tissues[tissue].vol.entities
        return np.hstack([gmsh.model.mesh.getElements(dim, e)[1] for e in entities])

    _get_tissue_surf_elems = partialmethod(_get_tissue_elems, dim=2)
    _get_tissue_vol_elems = partialmethod(_get_tissue_elems, dim=3)

    def _get_tissue_elems_coords(self, tissue, dim):
        """Return the barycenters of all the elements of the tissue in specified
        dimension.
        """
        if dim == 2:
            entities = self.tissues[tissue].surf.entities
            elems_type = 2  # Triangle
        elif dim == 3:
            entities = self.tissues[tissue].vol.entities
            elems_type = 4  # Tetrahedron
        return np.hstack(
            [
                gmsh.model.mesh.getBarycenters(elems_type, e, fast=False, primary=False)
                for e in entities
            ]
        ).reshape((-1, 3))

    _get_tissue_surf_elems_coords = partialmethod(_get_tissue_elems_coords, dim=2)
    _get_tissue_vol_elems_coords = partialmethod(_get_tissue_elems_coords, dim=3)

    def _fill_empty_field_elems(self, tissue, elems_tags, elems_vals, n_vals, fill_val):
        """Add a default value to empty elements."""
        all_elems_tags = self._get_tissue_vol_elems(tissue)
        empty_elems_idx = np.isin(
            all_elems_tags, elems_tags, assume_unique=True, invert=True
        )
        if np.any(empty_elems_idx):
            logger.debug(f"Filling {np.nonzero(empty_elems_idx).size} elements.")
            empty_elems_tags = all_elems_tags[empty_elems_idx]
            empty_elems_vals = np.broadcast_to(
                np.array(fill_val), (1, int(empty_elems_tags.size * n_vals))
            ).ravel()
            elems_tags = np.hstack((elems_tags.ravel(), empty_elems_tags))
            elems_vals = np.hstack((elems_vals.ravel(), empty_elems_vals))
        return elems_tags, elems_vals

    def _add_field_view(self, name, tissue, elems_tags, elems_vals, n_vals):
        """Add a view with the field values."""
        model = gmsh.model.list()[0]
        view = gmsh.view.add(f"{tissue}_{name}")
        elems_vals = elems_vals.reshape((-1, n_vals))
        logger.debug(f"{elems_tags.shape}, {elems_vals.shape}, {n_vals}")
        gmsh.view.addModelData(
            view, 0, model, "ElementData", elems_tags.ravel(), elems_vals
        )
        return view

    def _get_field_coords(self, shape, affine):
        """Convert cell indices into real coordinates."""
        flat_idx = np.arange(np.prod(shape[:3]))
        idx = np.vstack(np.unravel_index(flat_idx, shape[:3])).T
        coords = nib.affines.apply_affine(affine, idx)
        return coords

    def _interp_field(self, tissue, coords, field, method, fill_val):
        """Interpolate a field on a mesh grid."""
        x = np.unique(coords[:, 0])
        y = np.unique(coords[:, 1])
        z = np.unique(coords[:, 2])
        interpolate = RegularGridInterpolator(
            (x, y, z), field, method=method, bounds_error=False, fill_value=fill_val
        )
        gmsh.initialize()
        gmsh.open(str(self.mesh_path))
        elems_tags = self._get_tissue_vol_elems(tissue)
        elems_coords = self._get_tissue_vol_elems_coords(tissue)
        gmsh.finalize()
        elems_vals = interpolate(elems_coords)
        return elems_tags, elems_vals



"""Implement `DistNormal` and `DistTruncNormal` classes."""
from .abc import DistABC

import chaospy as chaos


class DistNormal(DistABC):
    """A normal distribution.

    Parameters
    ----------
    mu : float
        The mean of the distribution.
    sigma : float
        The standard deviation of the distribution.
    """

    def __init__(self, mu, sigma):
        super().__init__("normal")
        self.update({"mu": mu, "sigma": sigma})

    @property
    def mu(self):
        """Return the mean of the distribution.

        mu : float
            The mean of the distribution.
        """
        return self["mu"]

    @property
    def sigma(self):
        """Return the standard deviation of the distribution.

        sigma : float
            The standard deviation of the distribution.
        """
        return self["sigma"]

    @property
    def dist(self):
        """Return the actual distribution.

        Returns
        -------
        chaospy.Normal
            The actual distribution.
        """
        return chaos.Normal(mu=self.mu, sigma=self.sigma)

    @property
    def uniform_dist(self):
        """Return a uniform distribution used for sampling.

        Returns
        -------
        chaospy.Uniform
            The uniform distribution.
        """
        return chaos.Uniform(
            lower=self.mu - 3 * self.sigma, upper=self.mu + 3 * self.sigma
        )

    @property
    def salib_name(self):
        """Return the name of the distribution in SALib.

        Returns
        -------
        str
            The name of the distribution in SALib.
        """
        return "norm"

    @property
    def salib_bounds(self):
        """Return the bounds of the distribution in SALib.

        Returns
        -------
        list [float]
            The bounds of the distribution in SALib.
        """
        return [self.mu, self.sigma]


class DistTruncNormal(DistABC):
    """A truncated normal distribution.

    Parameters
    ----------
    mu : float
        The mean of the distribution.
    sigma : float
        The standard deviation of the distribution.
    lower : float
        The lower bound of the distribution.
    upper : float
        The upper bound of the distribution.
    """

    def __init__(self, mu, sigma, lower, upper):
        super().__init__("trunc_normal")
        self.update({"mu": mu, "sigma": sigma, "lower": lower, "upper": upper})

    @property
    def mu(self):
        """Return the mean of the distribution.

        mu : float
            The mean of the distribution.
        """
        return self["mu"]

    @property
    def sigma(self):
        """Return the standard deviation of the distribution.

        sigma : float
            The standard deviation of the distribution.
        """
        return self["sigma"]

    @property
    def lower(self):
        """Return the lower bound of the distribution.

        Returns
        -------
        float
            The lower bound of the distribution.
        """
        return self["lower"]

    @property
    def upper(self):
        """Return the upper bound of the distribution.

        Returns
        -------
        float
            The upper bound of the distribution.
        """
        return self["upper"]

    @property
    def dist(self):
        """Return the actual distribution.

        Returns
        -------
        chaospy.TruncNormal
            The actual distribution.
        """
        return chaos.Normal(
            mu=self.mu, sigma=self.sigma, lower=self.lower, upper=self.upper
        )

    @property
    def uniform_dist(self):
        """Return a uniform distribution used for sampling.

        Returns
        -------
        chaospy.Uniform
            The uniform distribution.
        """
        return chaos.Uniform(lower=self.lower, upper=self.upper)

    @property
    def salib_name(self):
        """Return the name of the distribution in SALib.

        Returns
        -------
        str
            The name of the distribution in SALib.
        """
        return "truncnorm"

    @property
    def salib_bounds(self):
        """Return the bounds of the distribution in SALib.

        Returns
        -------
        list [float]
            The bounds of the distribution in SALib.
        """
        return [self.lower, self.upper, self.mu, self.sigma]



"""API for `shamo.core.problems.single`."""
from .abc import ProbABC
from .getdp import ProbGetDP

from .components.abc import CompABC
from .components.file_path import CompFilePath
from .components.grid_sampler import CompGridSampler
from .components.sensors import CompSensors
from .components.tissue_property import CompTissueProp
from .components.tissues import CompTissues


__all__ = [
    "ProbABC",
    "ProbGetDP",
    "CompABC",
    "CompFilePath",
    "CompGridSampler",
    "CompSensors",
    "CompTissueProp",
    "CompTissues",
]



"""Implement `SolEEGLeadfield` class."""
import h5py
import numpy as np

from shamo.core.problems.single import CompSensors
from shamo.core.solutions.single import SolGetDP


class SolEEGLeadfield(SolGetDP):
    """Store information about an EEG leadfield matrix.

    Parameters
    ----------
    name : str
        The name of the solution.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the solution.

    Other Parameters
    ----------------
    sigmas : dict [str, list [float, str]]
        The electrical conductivity of the tissues.
    model_json_path : str
        The relative path of the model JSON file.
    markers : list [str]
        The names of the markers.
    reference : list [str]
        The name of the reference.
    rois : list [str]
        The names of the tissues of the region of interest.
    sensors : list [str]
        The names of the active sensors.
    shape : tuple [int]
        The shape of the matrix.
    use_grid : bool
        If ``True``, the source space is based on a grid.
    """

    def __init__(self, name, parent_path, **kwargs):
        super().__init__(name, parent_path, **kwargs)
        self.update(
            {
                "markers": kwargs.get("markers", []),
                "reference": kwargs.get("reference", []),
                "rois": kwargs.get("rois", []),
                "sensors": kwargs.get("sensors", []),
                "shape": tuple(kwargs.get("shape", [])),
                "use_grid": kwargs.get("use_grid", False),
            }
        )

    @property
    def matrix_path(self):
        """Return the path to the matrix HDF5 file.

        Returns
        -------
        pathlib.Path
            The path to the matrix HDF5 file.
        """
        return self.path / f"{self.name}.hdf5"

    @property
    def shape(self):
        """Return the shape of the matrix.

        Returns
        -------
        tuple [int]
            The shape of the matrix.
        """
        return self["shape"]

    @property
    def n_sensors(self):
        """Return the number of active sensors.

        Returns
        -------
        int
            The number of active sensors.
        """
        return self["shape"][0]

    @property
    def n_sources(self):
        """Return the number of sources.

        Returns
        -------
        int
            The number of sources.
        """
        return self["shape"][1]

    @property
    def markers(self):
        """Return the names of the markers.

        Returns
        -------
        list [str]
            The names of the markers.
        """
        return self["markers"]

    @property
    def reference(self):
        """Return the name of the reference.

        Returns
        -------
        str
            The name of the reference.
        """
        return self["reference"][0]

    @property
    def sensors(self):
        """Return the names of the active sensors.

        Returns
        -------
        list [str]
            The names of the active sensors.
        """
        return self["sensors"]

    @property
    def rois(self):
        """Return the names of the tissues of the region of interest.

        Returns
        -------
        list [str]
            The names of the tissues of the region of interest.
        """
        return self["rois"]

    @property
    def use_grid(self):
        """Return ``True`` if the source space is based on a grid.

        Returns
        -------
        bool
            ``True`` if the source space is based on a grid, ``False`` otherwise.
        """
        return self["use_grid"]

    @property
    def source_sp_path(self):
        """Return the path to the source space file.

        Returns
        -------
        pathlib.Path
            The path to the source space file.
        """
        if self.use_grid:
            return self.path / f"{self.name}_mask.nii"
        return self.path / f"{self.name}_elems.npz"

    def set_matrix(self, matrix):
        """Set the matrix of the solution.

        Parameters
        ----------
        numpy.ndarray
            The leadfield matrix.
        """
        with h5py.File(self.matrix_path, "w") as f:
            data = f.create_dataset(
                "e_field", matrix.shape, dtype="f", compression="lzf"
            )
            data[...] = matrix
        self["shape"] = matrix.shape

    def get_matrix(self):
        """Return the matrix.

        Returns
        -------
        h5py.Dataset
            The dataset containing the matrix.
        """
        return h5py.File(self.matrix_path, "r")["e_field"]



"""API for `shamo.core.fem`."""
from .group import Group
from .field import Field
from .tissue import Tissue
from .sensors.sensor import Sensor
from .sensors.point import PointSensor

from .fem import FEM

__all__ = ["FEM", "Field", "Group", "Tissue", "Sensor", "PointSensor"]



"""Implement the `Tissue` class."""
from . import Group, Field


class Tissue(dict):
    """A FEM tissue.

    Parameters
    ----------
    surf : dict [str, int or Iterable [int]] or shamo.fem.Group
        The surface physical group of the tissue.
    vol : dict [str, int or Iterable [int]] or shamo.fem.Group
        The volume physical group of the tissue.
    fields : dict [str, *]
        The fields defined in th tissue.
    """

    def __init__(self, surf, vol, fields={}):
        super().__init__(
            {
                "surf": Group(**surf),
                "vol": Group(**vol),
                "fields": {n: Field(**f) for n, f in fields.items()},
            }
        )

    @property
    def surf(self):
        """Return the surface group of the tissue.

        Returns
        -------
        shamo.fem.Group
            The surface group of the tissue.
        """
        return self["surf"]

    @property
    def vol(self):
        """Return the volume group of the tissue.

        Returns
        -------
        shamo.fem.Group
            The volume group of the tissue.
        """
        return self["vol"]

    @property
    def fields(self):
        """Return the fields of the tissue.

        Returns
        -------
        dict [str, shamo.fem.Field]
            The fields of the tissue.
        """
        return self["fields"]



"""Implement `SurrScalar` class."""
import json

from SALib.sample import saltelli
from SALib.analyze import sobol

from .abc import SurrABC


class SurrScalar(SurrABC):
    """Generate a Gaussian process from a set of scalar training data.

    Parameters
    ----------
    name : str
        The name of the generated surrogate model.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the surrogate model.

    Other Parameters
    ----------------
    params : list [tuple [str, shamo.DistABC]]
        A list of tuples containing the names of the parameters and the
        corresponding distributions as values.
    sol_json_path : str
        The path to the parametric solution the surrogate is built of.

    See Also
    --------
    shamo.core.surrogate.SurrABC
    """

    def __init__(self, name, parent_path, **kwargs):
        super().__init__(name, parent_path, **kwargs)
        self.update({"is_sobol_available": kwargs.get("is_sobol_available", False)})

    @property
    def is_sobol_available(self):
        """Return whether the Sobol indices are already computed.

        Returns
        -------
        bool
            Return ``True`` if the indices are already computed, ``False`` otherwise.
        """
        return self["is_sobol_available"]

    @property
    def sobol_path(self):
        """Return the path to the file storing the Sobol indices.

        Returns
        -------
        pathlib.Path
            The path to the file storing the Sobol indices.
        """
        return self.path / f"{self.name}_sobol.json"

    def get_sobol(self):
        """Return the precomputed Sobol indices.

        Returns
        -------
        dict [str, list [str]|dict [str, float]]|None
            A dictionary containing the names of the parameters, the values of the
            first, second and total order Sobol indices and the corresponding
            confidence with respect to the `conf_level`.
            If the file does not exist, returns ``None``.
        """
        if self.is_sobol_available:
            with open(self.sobol_path, "r") as f:
                return json.load(f)
        return None

    def gen_sobol(self, n=1000, n_resamples=100, conf_level=0.95):
        """Compute Sobol sensitivity indices.

        Parameters
        ----------
        n : int, optional
            The number of sample to predict with the Gaussian process. The default is
            ``1000``.
        n_resamples : int, optional
            The number of resamples. The default is ``100``.
        conf_level : float, optional
            The confidence interval level. The default is ``0.95``.

        Returns
        -------
        dict [str, list [str]|dict [str, float]]
            A dictionary containing the names of the parameters, the values of the
            first, second and total order Sobol indices and the corresponding
            confidence with respect to the `conf_level`.
        """
        indices = self.get_sobol()
        if indices is not None:
            return sobol
        params_names = [n for n, _ in self.params]
        prob = {
            "num_vars": len(self.params),
            "names": params_names,
            "dists": [d.salib_name for _, d in self.params],
            "bounds": [d.salib_bounds for _, d in self.params],
        }
        x = saltelli.sample(prob, n)
        y, _ = self.predict(x)
        s_i = sobol.analyze(prob, y, num_resamples=n_resamples, conf_level=conf_level)
        indices = {
            "params": params_names,
            "s1": {"val": s_i["S1"].tolist(), "conf": s_i["S1_conf"].tolist()},
            "s2": {"val": s_i["S2"].tolist(), "conf": s_i["S2_conf"].tolist()},
            "st": {"val": s_i["ST"].tolist(), "conf": s_i["ST_conf"].tolist()},
        }
        with open(self.sobol_path, "w") as f:
            json.dump(
                indices, f, indent=4, sort_keys=True,
            )
        return indices



"""Implement the `ProbEEGLeadfield` class."""
import logging
from pathlib import Path
from tempfile import TemporaryDirectory

import gmsh
import h5py
import nibabel as nib
import numpy as np
from scipy.sparse import csc_matrix

from shamo.core.problems.single import (
    ProbGetDP,
    CompFilePath,
    CompGridSampler,
    CompTissueProp,
    CompSensors,
    CompTissues,
)
from shamo.utils.onelab import (
    read_vector_file,
    get_elems_coords,
    get_elems_subset,
    pos_to_nii,
)
from .solution import SolEEGLeadfield

logger = logging.getLogger(__name__)


class ProbEEGLeadfield(ProbGetDP):
    """A problem definition to generate the EEG leadfield matrix.

    Attributes
    ----------
    sigmas : shamo.core.problems.single.CompTissueProp
        The electrical conductivity of the tissues.
    markers : shamo.core.problems.single.CompSensors
        The electrodes to ignore when solving the problem.
    reference : shamo.core.problems.single.CompSensors
        The electrode to use as the reference.
    rois : shamo.core.problems.single.CompTissues
        The tissues in which the leadfield must be computed.
    min_elems_dist : float
        The minimal distance between two elements in the region of interest.
    elems_path : shamo.core.problems.single.CompFilePath
        The path to the elements subset.
    grid : shamo.core.problems.single.CompGridSampler
        The grid sampler if the source space must be based on a grid.
    """

    _tmp_sort_idx = []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._source = CompSensors()
        self.markers = CompSensors()
        self.reference = CompSensors()
        self.rois = CompTissues()
        self.min_elems_dist = 0
        self.elems_path = CompFilePath()
        self.grid = CompGridSampler()

    @property
    def template(self):
        """Return the name of the template PRO file.

        Returns
        -------
        str
            The name of the template PRO file.
        """
        return "eeg_leadfield.tmplt"

    def solve(self, name, parent_path, model, **kwargs):
        """Solve the EEG forward problem and build the leadfield matrix.

        Parameters
        ----------
        name : str
            The name of the solution.
        parent_path : str, byte or os.PathLike
            The path to the parent directory of the solution.
        model : shamo.fem.FEM
            The finite element model to solve the problem for.

        Returns
        -------
        shamo.eeg.SolEEGLeadfield
            The EEG leafield object.
        """
        for n, t in model.tissues.items():
            self._vol.set(n, t.vol.group)
        sensors = list(model.sensors.keys())
        for s in sensors:
            if s not in self.reference["sensors"]:
                self._source["sensors"] = [s]
                break

        with TemporaryDirectory() as d:
            if self.min_elems_dist > 0:
                gmsh.initialize()
                gmsh.merge(str(model.mesh_path))
                entities = []
                for roi in self.rois["tissues"]:
                    entities.extend(model.tissues[roi].vol.entities)
                tags, coords = get_elems_subset(3, entities, self.min_elems_dist)
                gmsh.finalize()
                np.savez(Path(d) / "source_sp.npz", tags=tags, coords=coords)
                self.elems_path.set(Path(d) / "source_sp.npz")
            self._check_components(**model)
            sensors = self._gen_rhs(model, d)
            self._gen_pro_file(d, **kwargs, **model, active_sensors=sensors)
            self._run_getdp(model, d)
            src = Path(d) / f"{name}.hdf5"
            if self.elems_path.use_path:
                elems = np.load(self.elems_path.path)
                source_sp = [elems["tags"], elems["coords"]]
            with h5py.File(src, "w") as f:
                for i, s in enumerate(sensors):
                    if i == 0 and not self.elems_path.use_path:
                        row, source_sp = self._gen_row(i, d, model)
                    else:
                        row, _ = self._gen_row(i, d, model, source_sp)
                    if i == 0:
                        shape = (len(sensors), row.size)
                        data = f.create_dataset(
                            "e_field", shape, dtype="f", compression="lzf",
                        )
                    data[i, :] = row
            sol = SolEEGLeadfield(
                name,
                parent_path,
                markers=self.markers["sensors"],
                reference=self.reference["sensors"],
                rois=self.rois["tissues"],
                sensors=sensors,
                shape=shape,
                sigmas=self.sigmas,
                use_grid=self.grid.use_grid,
            )
            sol["model_json_path"] = str(sol.get_relative_path(model.json_path))
            Path.rename(src, sol.matrix_path)
            if self.grid.use_grid:
                source_sp.to_filename(sol.source_sp_path)
            else:
                np.savez(sol.source_sp_path, tags=source_sp[0], coords=source_sp[1])
            sol.save()
        return sol

    def _gen_row(self, i, tmp_dir, model, source_sp=None):
        """Generate a single row of the leadfield matrix.

        Parameters
        ----------
        i : int
            The index of the row.
        tmp_dir : str, byte or os.PathLike
            The path to the temporary directory.
        model : shamo.FEM
            The finite element model.
        source_sp : list [numpy.ndarray], optional
            The source space. The default is ``None``.

        Returns
        -------
        numpy.ndarray
            The row of the matrix.
        source_sp : nibabel.Nifti1Image|tuple [numpy.ndarray]
            The source space.
        """
        tmp_dir = Path(tmp_dir)
        if self.grid.use_grid:
            img = self.grid.nii_from_pos(tmp_dir / f"{i}.pos", tmp_dir / f"{i}.nii")
            row = img.get_fdata()[self.grid.mask].ravel()
            if i == 0:
                source_sp = nib.Nifti1Image(
                    self.grid.mask.astype(np.uint8), self.grid.affine
                )
        else:
            elem_type, elems_tags, row = read_vector_file(tmp_dir / f"{i}.e")
            if i == 0 and not self.elems_path.use_path:
                logger.info("Acquiring elements coordinates.")
                gmsh.initialize()
                gmsh.merge(str(model.mesh_path))
                coords = get_elems_coords(elem_type, elems_tags)
                gmsh.finalize()
                source_sp = (elems_tags, coords)
            elif self.elems_path.use_path:
                row = self._get_row_for_elems(i, elems_tags, row, source_sp)
            row = row.ravel()
        return row, source_sp

    def _get_row_for_elems(self, i, elems_tags, row, source_sp):
        """Get a subset of the values.

        Parameters
        ----------
        i : int
            The iteration index.
        elems_tags : numpy.ndarray
            The elements tags.
        row : numpy.ndarray
            The elements values.
        source_sp : list [numpy.ndarray]
            The subset of elements to get the values for.

        Returns
        -------
        numpy.ndarray
            The subset of values.
        """
        if i == 0:
            mask = np.in1d(elems_tags, source_sp[0], assume_unique=True)
            idx_tags = np.argsort(elems_tags[mask])
            idx_sp = np.argsort(source_sp[0])
            self._tmp_sort_idx = [mask, idx_tags, idx_sp]
        row = row[self._tmp_sort_idx[0]]
        new_row = np.zeros((source_sp[0].size, 3))
        new_row[self._tmp_sort_idx[2]] = row[self._tmp_sort_idx[1]]
        return new_row

    def _check_components(self, **kwargs):
        """Check if the components are properly set."""
        super()._check_components(**kwargs)
        self._source.check("source", **kwargs)
        self.reference.check("reference", **kwargs)
        self.rois.check("region of interest", **kwargs)
        self.elems_path.check("elements path", **kwargs)
        self.grid.check("grid", **kwargs)
        if self.elems_path.use_path and self.grid.use_grid:
            raise RuntimeError(
                "Both 'elems_path' and 'grid' are set. Only one of them is allowed."
            )

    def _prepare_pro_file_params(self, **kwargs):
        """Return the parameters required to generate the PRO file.

        Returns
        -------
        dict [str, ...]
            The parameters required to generate the PRO file.
        """
        params = super()._prepare_pro_file_params(**kwargs)
        params.update(
            {
                "rois": self.rois.to_pro_param(**kwargs),
                "sources": self._source.to_pro_param(**kwargs),
                "sinks": self.reference.to_pro_param(**kwargs),
                "n_sensors": len(kwargs.get("active_sensors", {})) - 1,
                "use_grid": self.grid.to_pro_param(**kwargs),
            }
        )
        return params

    def _prepare_py_file_params(self, **kwargs):
        """Return the parameters required to generate the PY file.

        Returns
        -------
        dict [str, ...]
            The parameters required to generate the PY file.
        """
        params = super()._prepare_py_file_params(**kwargs)
        params.update(
            {
                "rois": self.rois.to_py_param(**kwargs),
                "reference": self.reference.to_py_param(**kwargs),
                "markers": self.markers.to_py_param(**kwargs),
                "use_elems_path": self.elems_path.use_path,
                "use_grid": self.grid.use_grid,
            }
        )
        if self.elems_path.use_path:
            params["elems_path"] = self.elems_path.to_py_param(**kwargs)
        if self.grid.use_grid:
            params["grid"] = self.grid.to_py_param(**kwargs)
        return params

    def _gen_rhs(self, model, tmp_dir):
        """Generate the right hand sides for the problem.

        Parameters
        ----------
        model : shamo.FEM
            The finite element model.
        tmp_dir : str, byte or os.PathLike
            The path to the temporary directory.

        Returns
        -------
        list [str]
            The names of the active sensors.
        """
        logger.info("Generating right hand sides.")
        gmsh.initialize()
        gmsh.open(str(model.mesh_path))
        n_nodes = gmsh.model.mesh.getNodes()[0].size
        gmsh.finalize()
        ref_row_idx = model.sensors[self.reference["sensors"][0]].node - 1
        sensors = []
        i = 0
        for n, s in model.sensors.items():
            if n not in self.reference["sensors"] and n not in self.markers["sensors"]:
                rhs = csc_matrix(
                    ((1, -1), ((s.node - 1, ref_row_idx), (0, 0))),
                    shape=(n_nodes, 1),
                    dtype=np.int,
                )
                np.savetxt(Path(tmp_dir) / f"{i}.rhs", rhs.toarray(), fmt="%d")
                sensors.append(n)
                i += 1
        return sensors



"""Implement `SurrEEGLeadfield` and `SurrEEGLeadfieldDifToRef` classes."""
import shutil

import numpy as np

from shamo.core.surrogate import SurrABC, SurrScalar
from shamo.eeg import SolEEGLeadfield, SolParamEEGLeadfield


class SurrEEGLeadfield(SurrABC):
    """Provide a way to generate any leadfield matrix from a set of conductivity.

    Parameters
    ----------
    name : str
        The name of the generated surrogate model.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the surrogate model.

    Other Parameters
    ----------------
    params : list [tuple [str, shamo.DistABC]]
        A list of tuples containing the names of the parameters and the
        corresponding distributions as values.
    sol_json_path : str
        The path to the parametric solution the surrogate is built of.

    See Also
    --------
    shamo.core.surrogate.SurrABC
    """

    @classmethod
    def _get_data(cls, sol, **kwargs):
        """Extract relevant data from a parametric solution.

        Parameters
        ----------
        sol : shamo.core.solutions.parametric.SolParamABC
            The parametric solution to generate a surrogate model for.

        Returns
        -------
        numpy.ndarray
            The coordinates of the training points in the parameter space. Each row
            represents an evaluation.
        numpy.ndarray
            The observations of the actual model at each coordinate from `x`. Each row
            represents an observation.
        list [tuple [str, shamo.DistABC]]
            A list of tuples containing the names of the parameters and the
            corresponding distributions as values.

        See Also
        --------
        shamo.core.surrogate.SurrABC
        """
        params = [
            [t, d[0]] for t, d in sol.sigmas.items() if d[0].dist_type != "constant"
        ]
        sols = sol.get_sub_sols()
        x = []
        y = []
        for s in sols:
            x.append([s.sigmas[t][0] for t, _ in params])
            y.append(np.array(s.get_matrix()).ravel())
        x = np.array(x)
        y = np.array(y)
        return x, y, params

    @classmethod
    def _check_params(cls, **kwargs):
        """Check if the parameters are properly set."""
        pass

    def _post_pro(self, x, y_mean, y_std, **kwargs):
        """Reshape the vectors into matrices.

        Parameters
        ----------
        numpy.ndarray
            The evaluation points.
        numpy.ndarray
            The evaluations mean.
        numpy.ndarray
            The evaluations standard deviations.

        Returns
        -------
        numpy.ndarray
            The evaluations mean.
        numpy.ndarray
            The evaluations standard deviations.
        """
        sol = SolParamEEGLeadfield.load(self.sol_json_path)
        shape = (x.shape[0], *sol.shape)
        y_mean_post = np.zeros(shape)
        for i in range(shape[0]):
            y_mean_post[i, :, :] = y_mean[i, :].reshape(shape[1:])
        return y_mean_post, y_std

    def predict_sol(self, name, parent_path, x, skip=0, **kwargs):
        """Generate new `SolEEGLeadfield` by evaluating the Gaussian process.

        Parameters
        ----------
        name : str
            The base name of the generated solution.
        parent_path : str, byte or os.PathLike
            The path to the parent directory of the solution.
        x : numpy.ndarray
            The evaluation points.
        skip : int, optional
            The starting index of the names of the solutions. The default is ``0``.

        Returns
        -------
        list [shamo.eeg.SolEEGLeadfield]
            The generated solutions.
        """
        param_sol = SolParamEEGLeadfield.load(self.sol_json_path)
        sol = SolEEGLeadfield.load(param_sol.sub_json_paths[0])
        sigmas = {t: s for t, s in sol.sigmas.items()}
        y, _ = self.predict(x, **kwargs)
        sols = []
        for i in range(x.shape[0]):
            for j, (t, d) in enumerate(self.params):
                sigmas[t][0] = x[i][j]
            s = SolEEGLeadfield(
                f"{name}_{i + skip:08d}",
                parent_path,
                markers=param_sol.markers,
                reference=param_sol.reference,
                rois=param_sol.rois,
                sensors=param_sol.sensors,
                shape=param_sol.shape,
                sigmas=sigmas,
                use_grid=param_sol.use_grid,
            )
            s["model_json_path"] = str(s.get_relative_path(sol.model_json_path))
            s.set_matrix(y[i, :, :].squeeze())
            shutil.copy(sol.source_sp_path, s.source_sp_path)
            s.save()
            sols.append(s)
        return sols


class SurrEEGLeadfieldToRef(SurrScalar):
    """Provide a way to evaluate the sensitivity of the EEG leadfield.

    Parameters
    ----------
    name : str
        The name of the generated surrogate model.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the surrogate model.

    Other Parameters
    ----------------
    params : list [tuple [str, shamo.DistABC]]
        A list of tuples containing the names of the parameters and the
        corresponding distributions as values.
    sol_json_path : str
        The path to the parametric solution the surrogate is built of.

    See Also
    --------
    shamo.core.surrogate.SurrScalar
    """

    @classmethod
    def _get_data(cls, sol, metric=None, **kwargs):
        """Extract relevant data from a parametric solution.

        Parameters
        ----------
        sol : shamo.core.solutions.parametric.SolParamABC
            The parametric solution to generate a surrogate model for.
        metric : function
            A function taking two parameters. First, the reference matrix and second the
            leadfield matrix considered. It must return a single scalar.

        Returns
        -------
        numpy.ndarray
            The coordinates of the training points in the parameter space. Each row
            represents an evaluation.
        numpy.ndarray
            The observations of the actual model at each coordinate from `x`. Each row
            represents an observation.
        list [tuple [str, shamo.DistABC]]
            A list of tuples containing the names of the parameters and the
            corresponding distributions as values.
        """
        ref = kwargs.get("ref", None)
        m_ref = np.array(ref.get_matrix())
        params = [
            [t, d[0]] for t, d in sol.sigmas.items() if d[0].dist_type != "constant"
        ]
        sols = sol.get_sub_sols()
        x = []
        y = []
        for s in sols:
            x.append([s.sigmas[t][0] for t, _ in params])
            m = s.get_matrix()
            y.append(metric(m_ref, m))
        x = np.array(x).reshape((len(sols), -1))
        y = np.array(y).reshape((len(sols),))
        return x, y, params

    @classmethod
    def _check_params(cls, **kwargs):
        """Check if the parameters are properly set.

        Raises
        ------
        RuntimeError
            If the argument `ref` is not given.
        """
        if kwargs.get("ref", None) is None:
            raise RuntimeError("Missing 'ref' argument.")



"""Implement the `Group` class."""
from collections.abc import Iterable
import logging

logger = logging.getLogger(__name__)


class Group(dict):
    """A FEM physical group.

    Parameters
    ----------
    dim : int
        The dimension of the group. Can be one of `DIM_POINT`, `DIM_LINE`, `DIM_SURF` or
        `DIM_VOL`.
    entities : int or Iterable [int]
        The tags of the entities composing the physical group.
    group : int
        The physical group tag.

    Raises
    ------
    ValueError
        If argument `dim` is not convertible to `int`.
        If an element of argument `entities` is not convertible to `int`.
        If argument `group` is not convertible to `int`.
    """

    DIM_POINT = 0
    DIM_LINE = 1
    DIM_SURF = 2
    DIM_VOL = 3

    def __init__(self, dim, entities, group):
        if not isinstance(entities, Iterable):
            entities = [entities]
        super().__init__(
            {"dim": int(dim), "entities": entities, "group": int(group),}
        )

    @property
    def dim(self):
        """Return the dimension of the physical group.

        Returns
        -------
        int
            The dimension of the physical group.
        """
        return self["dim"]

    @property
    def entities(self):
        """Return the tags of the entities composing the physical group.

        Returns
        -------
        set [int]
            The tags of the entities composing the physical group.
        """
        return self["entities"]

    @property
    def n_entities(self):
        """Return the number of entities composing the physical group.

        Returns
        -------
        int
            The number of entities composing the physical group.
        """
        return len(self.entities)

    @property
    def group(self):
        """Return the tag of the physical group.

        Returns
        -------
        int
            The tag of the physical group.
        """
        return self["group"]



"""Implement `ProbParamABC` class."""
from abc import abstractmethod, abstractproperty
import itertools as iter
import logging
import multiprocessing as mp
from pathlib import Path
import re

import chaospy as chaos
from jinja2 import Environment, PackageLoader
import numpy as np

from shamo.core.problems.single import ProbABC

logger = logging.getLogger(__name__)


class ProbParamABC(ProbABC):
    """A base class for any parametric problem."""

    METHOD_SEQ = "sequential"
    METHOD_MUL = "multiprocessing"
    METHOD_JOB = "job"

    @abstractproperty
    def template(self):
        """Return the path to the template PY file."""

    @abstractmethod
    def _gen_fixed_varying(self, **kwargs):
        """Generate two lists containing the fixed and varying parameters.

        Returns
        -------
        list [dict {str, ...}]
            The fixed parameters.
        list [dict {str, ...}]
            The varying parameters.
        """

    def _gen_params(self, n_evals, skip=0, **kwargs):
        """Generate the evaluation points.

        Parameters
        ----------
        n_evals : int
            The number of points to generate.
        skip : int, optional
            The number of points to skip at the beginning of the sequence. (The default
            is ``0``)

        Returns
        -------
        dict [str, dict[str, numpy.ndarray]]
            The evaluation points.

        Raises
        ------
        RuntimeError
            If no parameter is set as a random variable.

        Notes
        -----
        The parameters space is sampled by first modelling each random variable as a
        uniform distribution and then drawing the points from a Halton quasi-random
        sequence.

        This allows for a better covering of the whole space than if the actual
        distributions were used for each random input.
        """
        fixed, varying = self._gen_fixed_varying(**kwargs)
        if len(varying) == 0:
            raise RuntimeError(
                "No varying parameter was found. Use 'ProbEEGLeadfield' instead."
            )
        dist = chaos.J(*[p[0].uniform_dist for _, p in varying])
        x = dist.sample(n_evals + skip, rule="halton").reshape((len(dist), -1))
        if skip != 0:
            x = x[:, skip:]
        params = {}
        for i, (n, p) in enumerate(varying):
            prop, name = ProbParamABC._split_prop_name(n)
            if prop not in params:
                params[prop] = {}
            params[prop][name] = [x[i, :].ravel(), p[1]]
        for n, p in fixed:
            prop, name = ProbParamABC._split_prop_name(n)
            if prop not in params:
                params[prop] = {}
            params[prop][name] = [np.ones((x.shape[1],)) * p[0], p[1]]
        return params

    @staticmethod
    def _split_prop_name(name):
        """Split the names of the parameters."""
        match = re.match(r"^(?P<prop>[a-zA-Z]+)\.(?P<name>\w+)$", name)
        return match.group("prop"), match.group("name")

    @abstractmethod
    def _gen_sub_prob(self, i, **kwargs):
        """Generate one sub-problem.

        Parameters
        ----------
        i : int
            The index of the sub-problem.

        Returns
        -------
        shamo.core.problems.single.ProbABC
            The generated sub-problem.
        """

    @abstractmethod
    def _gen_sub_probs(self, n_evals, **kwargs):
        """Generate all the sub-problems.

        Parameters
        ----------
        n_evals : int
            The number of evaluation points.
        """
        return [self._gen_sub_prob(i, **kwargs) for i in range(n_evals)]

    def _solve_sub_prob(self, problem, name, parent_path, kwargs):
        """Solve one sub-problem.

        Parameters
        ----------
        problem : shamo.core.problems.single.ProbABC
            The sub-problem to solve.
        name : str
            The name of the sub-problem.
        parent_path : str, byte or os.PathLike
            The path to the parent directory of the sub-problem.
        kwargs : dict[str, ...]
            Any other arguments.

        Returns
        -------
        shamo.core.objects.ObjDir
            The solution to the problem.
        """
        return problem.solve(name, parent_path, **kwargs)

    def _gen_py_file(self, prob, name, parent_path, kwargs):
        """Generate a PY file to solve a sub-problem later.

        Parameters
        ----------
        prob : shamo.core.problems.single.ProbABC
            The problem to solve.
        name : str
            The name of the sub-problem.
        parent_path : str, byte or os.PathLike
            The path to the parent directory of the sub-problem.
        kwargs : dict[str, ...]
            Any other arguments.

        Returns
        -------
        list []
        """
        logger.info("Generating python files.")
        env = Environment(loader=PackageLoader("shamo", "templates/py"))
        template = env.get_template(self.template)
        content = template.render(
            problem=prob,
            name=name,
            parent_path=str(parent_path),
            **kwargs,
            **prob._prepare_py_file_params(**kwargs),
        )
        with open(Path(parent_path) / f"{name}.py", "w") as f:
            f.write(content)
        return []

    def _solve_sub_probs(
        self, sol, sub_probs, method="sequential", n_proc=1, **kwargs,
    ):
        """Solve the sub-problems.

        Parameters
        ----------
        sol : shamo.core.solutions.parametric.SolParamABC
            The parametric solution which the sub-solution will be part of.
        sub_probz : shamo.core.problems.single.ProbABC
            The sub-problemz to solve.
        method : str, optional
            The method to solve the sub-problems. The accepted values are
            ``'sequential'``, ``'multiprocessing'``, ``'job'``. (The default is
            ``'sequential'``)
        n_proc : int, optional
            The number of processes to solve the problem when `method` is set to
            ``'multiprocessing'``. (The default is ``1``)

        Notes
        -----
        If `method` is set to ``'job'``, the method will only produce PY files. Each of
        these scripts will solve a sub-problem. Once all the sub-problems are solved,
        one must call `shamo.core.solutions.parametric.SolABC.finalize()` to finalize
        the resolution.
        """
        logger.info(f"Solving {len(sub_probs)} sub-problems.")
        generator = (
            [p, f"{sol.name}_{i:08d}", sol.path, kwargs]
            for i, p in enumerate(sub_probs)
        )
        sub_sols = []
        if method == self.METHOD_SEQ:
            sub_sols = list(iter.starmap(self._solve_sub_prob, generator))
            sol.finalize(**kwargs)
        elif method == self.METHOD_MUL:
            with mp.Pool(processes=n_proc) as p:
                sub_sols = list(p.starmap(self._solve_sub_prob, generator))
            sol.finalize(**kwargs)
        else:
            sub_sols = list(iter.starmap(self._gen_py_file, generator))
        return sub_sols

    @abstractmethod
    def solve(
        self,
        name,
        parent_path,
        n_evals,
        method="sequential",
        n_proc=1,
        skip=0,
        **kwargs,
    ):
        """Solve the parametric problem.

        Parameters
        ----------
        name : str
            The name of the solution.
        parent_path : str, byte or os.PathLike
            The path to the parent directory of the solution.
        n_evals : int
            The number of evaluation points.
        method : str, optional
            The method to solve the sub-problems. The accepted values are
            ``'sequential'``, ``'multiprocessing'``, ``'job'``. (The default is
            ``'sequential'``)
        n_proc : int, optional
            The number of processes to solve the problem when `method` is set to
            ``'multiprocessing'``. (The default is ``1``)
        skip : int, optional
            The number of points to skip at the beginning of the sequence. (The default
            is ``0``)

        Returns
        -------
        shamo.core.solutions.parametric.SolParamABC
            The generated solution.
        """



"""API for `shamo`."""
import warnings

from shamo.core.distributions import *
from shamo.core.fem import *

__version__ = "1.0.0"

# Remove unnecessary warnings
warnings.filterwarnings("ignore", category=RuntimeWarning, lineno=521)



"""Implement the `Field` class."""
import logging

logger = logging.getLogger(__name__)


class Field(dict):
    """A FEM field.

    Parameters
    ----------
    field_type : str
        The type of the field.
    view : int
        The view corresponding to the field.
    formula : str, optional
        The formula linking the field to a physical property.

    Notes
    -----
    The argument `formula` must be a valid formatter expression with named placeholders.
    To reference a tissue property, use the dict accessor.
    Available properties are:
        - 'sigma[tissue]'
    """

    SCALAR = "scalar"
    VECTOR = "vector"
    TENSOR = "tensor"

    def __init__(self, field_type, view, formula="1"):
        field_type = str(field_type)
        if field_type not in {self.SCALAR, self.VECTOR, self.TENSOR}:
            raise ValueError(
                "Argument 'field_type' must be one of 'scalar', 'vector' or 'tensor'."
            )
        super().__init__(
            {"field_type": field_type, "view": int(view), "formula": formula}
        )

    @property
    def field_type(self):
        """Return the type of the fied.

        Returns
        -------
        str
            The type of the field.
        """
        return self["field_type"]

    @property
    def view(self):
        """Return the view corresponding to the field.

        Returns
        -------
        int
            The view corresponding to the field.
        """
        return self["view"]

    @property
    def formula(self):
        """Return the formula linking the field to a physical property.

        Returns
        -------
        str
            The formula linking the field to a physical property.
        """
        return self["formula"]

    def gen_formula(self, **kwargs):
        """Generate the formula to use the field in a property.

        Returns
        -------
        str
            The generated formula.
        """
        formula = self.formula.format(**kwargs)
        field_type = f"{self.field_type.capitalize()}Field"
        return f"{formula} * {field_type}[XYZ[]]{{ {self.view} }}"
